package de.ozgcloud.bescheid;

import java.util.Optional;

import jakarta.annotation.PostConstruct;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.boot.info.BuildProperties;
import org.springframework.stereotype.Service;

import de.ozgcloud.bescheid.attacheditem.AttachedItem;
import de.ozgcloud.bescheid.attacheditem.AttachedItemService;
import de.ozgcloud.bescheid.attributes.ClientAttributeService;
import de.ozgcloud.bescheid.common.user.UserProfileService;
import de.ozgcloud.bescheid.nachricht.NachrichtService;
import de.ozgcloud.bescheid.vorgang.Vorgang;
import de.ozgcloud.bescheid.vorgang.VorgangId;
import de.ozgcloud.bescheid.vorgang.VorgangService;
import de.ozgcloud.common.binaryfile.FileId;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.document.Document;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
@RequiredArgsConstructor
class BescheidService {

	private static final String ERROR_MESSAGE_NO_SERVICE = "No Bescheid Endpoint is configured.";

	private final VorgangService vorgangService;
	private final AttachedItemService attachedItemService;
	private final NachrichtService nachrichtService;
	private final UserProfileService currentUserService;
	private final ClientAttributeService bescheidClientAttributeService;
	private final BuildProperties buildProperties;
	private final Optional<BescheidRemoteService> remoteService;

	@PostConstruct
	void logStatus() {
		remoteService.ifPresentOrElse(
				service -> LOG.info("Bescheid-Manager is configured."),
				() -> LOG.info("No BescheidRemoteService configured - Bescheid creation is not possible."));
	}

	public Bescheid createBescheid(BescheidRequest request) {
		checkRemoteService();

		return doCreateBescheid(request);
	}

	private Bescheid doCreateBescheid(BescheidRequest request) {
		var vorgang = vorgangService.getById(request.getVorgangId());
		return remoteService.map(service -> service.create(request, vorgang))
				.map(bescheid -> updateBescheid(bescheid, vorgang))
				.orElseThrow(() -> new TechnicalException(ERROR_MESSAGE_NO_SERVICE));
	}

	private Bescheid updateBescheid(Bescheid bescheid, Vorgang vorgang) {
		return bescheid.toBuilder().vorgangId(vorgang.getId()).serviceKonto(vorgang.getServiceKonto()).build();
	}

	private void checkRemoteService() {
		if (remoteService.isEmpty()) {
			LOG.error(ERROR_MESSAGE_NO_SERVICE);
			throw new TechnicalException(ERROR_MESSAGE_NO_SERVICE);
		}
	}

	public void sendBescheidManually(AttachedItem bescheidItem, long version) {
		validateBescheidSendManually(bescheidItem, version);
		sendBescheid(bescheidItem);
	}

	void validateBescheidSendManually(AttachedItem bescheidItem, long version) {
		validateBescheid(bescheidItem, version);
		var sendBy = bescheidItem.getItem().get(Bescheid.FIELD_SEND_BY);
		if (Bescheid.SendBy.MANUAL.notValue(sendBy)) {
			throw new TechnicalException("Bescheid has unexpected sendBy value: '%s'. Expected is %s".formatted(sendBy, Bescheid.SendBy.MANUAL));
		}
	}

	public void sendBescheidPostfachMail(AttachedItem bescheidItem, long version) {
		validateBescheidSendPostfach(bescheidItem, version);
		var vorgang = vorgangService.getById(VorgangId.from(bescheidItem.getVorgangId()));
		nachrichtService.sendNachricht(buildBescheid(bescheidItem, vorgang.getServiceKonto()));
		sendBescheid(bescheidItem);
	}

	public void validateBescheidSendPostfach(AttachedItem bescheidItem, long version) {
		validateBescheid(bescheidItem, version);
		var sendBy = bescheidItem.getItem().get(Bescheid.FIELD_SEND_BY);
		if (Bescheid.SendBy.NACHRICHT.notValue(sendBy)) {
			throw new TechnicalException("Bescheid has unexpected sendBy value: '%s'. Expected is %s".formatted(sendBy, Bescheid.SendBy.NACHRICHT));
		}
		if (StringUtils.isBlank(getNachrichtSubject(bescheidItem))) {
			throw new TechnicalException("Bescheid has no nachricht subject");
		}
		if (StringUtils.isBlank(getNachrichtText(bescheidItem))) {
			throw new TechnicalException("Bescheid has no nachricht text");
		}
	}

	void validateBescheid(AttachedItem bescheidItem, long version) {
		if (bescheidItem.getVersion() != version) {
			throw new TechnicalException("Bescheid has different version. Expected: %d, but was: %d".formatted(version, bescheidItem.getVersion()));
		}
		var status = MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_STATUS);
		if (Bescheid.Status.DRAFT.not(status)) {
			throw new TechnicalException("Bescheid has status '%s'. Bescheid must have status DRAFT".formatted(status));
		}
		if (StringUtils.isBlank(MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_BESCHEID_DOCUMENT))) {
			throw new TechnicalException("Bescheid has no document");
		}
	}

	Bescheid buildBescheid(AttachedItem bescheidItem, Vorgang.ServiceKonto serviceKonto) {
		return Bescheid.builder()
				.vorgangId(VorgangId.from(bescheidItem.getVorgangId()))
				.genehmigt(getBewilligt(bescheidItem))
				.bescheidFileId(getBescheidFileId(bescheidItem))
				.nachrichtSubject(Optional.ofNullable(getNachrichtSubject(bescheidItem)))
				.nachrichtText(Optional.ofNullable(getNachrichtText(bescheidItem)))
				.createdBy(currentUserService.getUserProfile().getId())
				.serviceKonto(serviceKonto)
				.build();
	}

	FileId getBescheidFileId(AttachedItem bescheidItem) {
		var bescheidDocument = attachedItemService.getItem(MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_BESCHEID_DOCUMENT));
		return FileId.from(MapUtils.getString(bescheidDocument.getItem(), Document.FIELD_DOCUMENT_FILE));
	}

	String getNachrichtSubject(AttachedItem bescheidItem) {
		return MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_NACHRICHT_SUBJECT);
	}

	String getNachrichtText(AttachedItem bescheidItem) {
		return MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_NACHRICHT_TEXT);
	}

	void sendBescheid(AttachedItem bescheidItem) {
		attachedItemService.setBescheidStatus(bescheidItem.getId(), bescheidItem.getVersion(), Bescheid.Status.SENT);
		try {
			vorgangService.bescheiden(bescheidItem.getVorgangId());
		} catch (Exception e) {
			attachedItemService.setBescheidStatus(bescheidItem.getId(), bescheidItem.getVersion(), Bescheid.Status.DRAFT);
			throw e;
		}
		bescheidClientAttributeService.setAntragResult(bescheidItem.getVorgangId(), getBewilligt(bescheidItem));
	}

	boolean getBewilligt(AttachedItem bescheidItem) {
		return MapUtils.getBooleanValue(bescheidItem.getItem(), Bescheid.FIELD_BEWILLIGT, false);
	}

	public BescheidManagerConfig getConfig() {
		return BescheidManagerConfig.builder()
				.version(buildProperties.getVersion())
				.javaVersion(getJavaVersion())
				.features(buildFeatures())
				.build();
	}

	String getJavaVersion() {
		return System.getProperty("java.version");
	}

	BescheidManagerConfig.Features buildFeatures() {
		return BescheidManagerConfig.Features.builder()
				.canCreateBescheidDocument(remoteService.isPresent())
				.build();
	}
}
