/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid.attacheditem;

import static de.ozgcloud.bescheid.attacheditem.AttachedItemService.*;

import java.time.LocalDate;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.collections.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.ozgcloud.bescheid.Bescheid;
import de.ozgcloud.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.bescheid.vorgang.VorgangId;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.VorgangAttachedItemServiceGrpc.VorgangAttachedItemServiceBlockingStub;
import io.grpc.ClientInterceptor;
import lombok.extern.log4j.Log4j2;
import net.devh.boot.grpc.client.inject.GrpcClient;

@Service
@Log4j2
class VorgangAttachedItemRemoteService {

	private static final Comparator<AttachedItem> BY_BESCHIEDEN_AM_DESC = (bescheid1, bescheid2) -> {
		var beschiedenAm1 = LocalDate.parse(MapUtils.getString(bescheid1.getItem(), Bescheid.FIELD_BESCHIEDEN_AM));
		var beschiedenAm2 = LocalDate.parse(MapUtils.getString(bescheid2.getItem(), Bescheid.FIELD_BESCHIEDEN_AM));
		return beschiedenAm2.compareTo(beschiedenAm1);
	};

	@GrpcClient("vorgang-manager")
	private VorgangAttachedItemServiceBlockingStub serviceStub;
	@Autowired
	private ClientInterceptor bescheidCallContextInterceptor;
	@Autowired
	private AttachedItemMapper attachedItemMapper;

	public Optional<AttachedItem> findBescheidDraft(String vorgangId) {
		return findBescheidDraft(buildFindRequest(vorgangId));
	}

	public List<AttachedItem> findAllBescheid(VorgangId vorgangId) {
		return findAllBescheid(buildFindRequest(vorgangId.toString())).toList();
	}

	GrpcFindVorgangAttachedItemRequest buildFindRequest(String vorgangId) {
		return GrpcFindVorgangAttachedItemRequest.newBuilder()
				.setVorgangId(vorgangId)
				.setItemName(BESCHEID_ITEM_NAME)
				.setClient(BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME)
				.build();
	}

	Optional<AttachedItem> findBescheidDraft(GrpcFindVorgangAttachedItemRequest request) {
		var bescheidItems = findAllBescheid(request).filter(this::hasStatusDraft).toList();
		if (bescheidItems.size() > 1) {
			LOG.error("Found more than one ({}) draft bescheid for vorgangId: {}. Return one with last beschiedenAm date", bescheidItems.size(),
					bescheidItems.getFirst().getVorgangId());
		}
		return bescheidItems.isEmpty() ? Optional.empty() : Optional.of(bescheidItems.getFirst());
	}

	Stream<AttachedItem> findAllBescheid(GrpcFindVorgangAttachedItemRequest request) {
		var grpcResponse = getServiceStub().find(request);
		return grpcResponse.getVorgangAttachedItemsList().stream().map(attachedItemMapper::mapFromVorgangAttachedItem)
				.sorted(BY_BESCHIEDEN_AM_DESC);
	}

	boolean hasStatusDraft(AttachedItem bescheidItem) {
		return Bescheid.Status.DRAFT.hasValue(bescheidItem.getItem().get(Bescheid.FIELD_STATUS));
	}

	public AttachedItem getBescheid(String bescheidId) {
		var grpcVorgangAttachedItemResponse = getServiceStub().getById(buildGetByIdRequest(bescheidId));
		return attachedItemMapper.mapFromVorgangAttachedItem(grpcVorgangAttachedItemResponse.getVorgangAttachedItem());
	}

	public AttachedItem getItem(String id) {
		var grpcVorgangAttachedItemResponse = getServiceStub().getById(buildGetByIdRequest(id));
		return attachedItemMapper.mapFromVorgangAttachedItem(grpcVorgangAttachedItemResponse.getVorgangAttachedItem());
	}

	GrpcVorgangAttachedItemRequest buildGetByIdRequest(String bescheidId) {
		return GrpcVorgangAttachedItemRequest.newBuilder().setId(bescheidId).build();
	}

	VorgangAttachedItemServiceBlockingStub getServiceStub() {
		return serviceStub.withInterceptors(bescheidCallContextInterceptor);
	}
}
