package de.ozgcloud.bescheid.nachricht;

import java.io.IOException;
import java.io.StringWriter;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.bescheid.Bescheid;
import de.ozgcloud.bescheid.vorgang.Vorgang.PostfachAddress;
import de.ozgcloud.common.errorhandling.TechnicalException;
import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
public class NachrichtService {

	public static final String SEND_POSTFACH_NACHRICHT_ORDER = "SEND_POSTFACH_NACHRICHT";
	public static final String FIELD_POSTFACH_ID = "postfachId";
	static final String FIELD_REPLY_OPTION = "replyOption";
	static final String REPLY_OPTION = "POSSIBLE";
	static final String FIELD_SUBJECT = "subject";
	static final String FIELD_MAIL_BODY = "mailBody";
	static final String FIELD_ATTACHMENTS = "attachments";

	@Autowired
	private NachrichtRemoteService remoteService;
	@Autowired
	private Configuration freemarkerCfg;
	@Autowired
	@Qualifier("bescheid_OzgCloudCommandService")
	private OzgCloudCommandService commandService;
	@Autowired
	private CommandMapper commandMapper;

	static final String SUBJECT = "Ihr Antrag";

	private static final String TEMPLATE_FILE = "bescheid.nachrichtTemplate.txt.ftlh";

	public void createNachrichtDraft(Bescheid bescheid) {
		buildNachricht(bescheid).ifPresentOrElse(remoteService::saveDraft, () -> LOG.warn("No ServiceKonto given on Vorgang."));
	}

	Optional<Nachricht> buildNachricht(Bescheid bescheid) {
		return getAddress(bescheid).map(address -> Nachricht.builder()
				.vorgangId(bescheid.getVorgangId())
				.postfachAddress(address)
				.subject(bescheid.getNachrichtSubject().orElse(SUBJECT))
				.mailBody(buildMessage(bescheid))
				.createdBy(bescheid.getCreatedBy())
				.bescheidFileId(bescheid.getBescheidFileId())
				.build());
	}

	String buildMessage(Bescheid bescheid) {
		return bescheid.getNachrichtText()
				.orElseGet(() -> fillTemplate(TEMPLATE_FILE, bescheid));
	}

	String fillTemplate(String templateName, Object dataModel) {
		try {
			Template template = getTemplate(templateName);
			StringWriter stringWriter = new StringWriter();
			template.process(dataModel, stringWriter);
			return stringWriter.toString();

		} catch (IOException | TemplateException e) {
			throw new TechnicalException("Error filling template", e);
		}
	}

	Template getTemplate(String templateName) {
		try {
			return freemarkerCfg.getTemplate(templateName);
		} catch (IOException e) {
			throw new TechnicalException("Error loading mail template", e);
		}
	}

	public void sendNachricht(Bescheid bescheid) {
		commandService.createAndWaitUntilDone(buildSendBescheidCommand(bescheid));
	}

	OzgCloudCommand buildSendBescheidCommand(Bescheid bescheid) {
		return OzgCloudCommand.builder()
				.vorgangId(commandMapper.toOzgCloudVorgangId(bescheid.getVorgangId().toString()))
				.relationId(commandMapper.mapRelationId(bescheid.getVorgangId().toString()))
				.order(SEND_POSTFACH_NACHRICHT_ORDER)
				.createdBy(commandMapper.toOzgCloudUserId(bescheid.getCreatedBy().toString()))
				.bodyObject(buildSendNachrichtCommandBody(bescheid))
				.build();
	}

	Map<String, Object> buildSendNachrichtCommandBody(Bescheid bescheid) {
		return Map.of(FIELD_POSTFACH_ID, getPostfachId(bescheid),
				FIELD_REPLY_OPTION, REPLY_OPTION,
				FIELD_SUBJECT, bescheid.getNachrichtSubject().orElse(SUBJECT),
				FIELD_MAIL_BODY, bescheid.getNachrichtText().orElse(StringUtils.EMPTY),
				FIELD_ATTACHMENTS, bescheid.getBescheidFileId().toString()
		);

	}

	String getPostfachId(Bescheid bescheid) {
		return getAddress(bescheid)
				.map(PostfachAddress::getIdentifier)
				.map(identifier -> MapUtils.getString(identifier, FIELD_POSTFACH_ID))
				.filter(StringUtils::isNotBlank)
				.orElseThrow(() -> new TechnicalException("No PostfachId found for Bescheid (vorgangId: %s)".formatted(bescheid.getVorgangId())));
	}

	Optional<PostfachAddress> getAddress(Bescheid bescheid) {
		var serviceKonto = bescheid.getServiceKonto();
		if (Objects.nonNull(serviceKonto)) {
			return Optional.of(serviceKonto.getPostfachAddresses().get(0));
		}
		return Optional.empty();
	}

}
