package de.ozgcloud.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.boot.info.BuildProperties;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.bescheid.attacheditem.AttachedItem;
import de.ozgcloud.bescheid.attacheditem.AttachedItemService;
import de.ozgcloud.bescheid.attacheditem.AttachedItemTestFactory;
import de.ozgcloud.bescheid.attributes.ClientAttributeService;
import de.ozgcloud.bescheid.common.user.UserProfile;
import de.ozgcloud.bescheid.common.user.UserProfileService;
import de.ozgcloud.bescheid.nachricht.NachrichtService;
import de.ozgcloud.bescheid.vorgang.ServiceKontoTestFactory;
import de.ozgcloud.bescheid.vorgang.Vorgang;
import de.ozgcloud.bescheid.vorgang.VorgangId;
import de.ozgcloud.bescheid.vorgang.VorgangService;
import de.ozgcloud.bescheid.vorgang.VorgangTestFactory;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.common.binaryfile.FileId;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.document.DocumentTestFactory;

class BescheidServiceTest {

	@Spy
	@InjectMocks
	private BescheidService service;

	@Mock
	private VorgangService vorgangService;
	@Mock
	private BescheidRemoteService remoteService;
	@Mock
	private AttachedItemService attachedItemService;
	@Mock
	private NachrichtService nachrichtService;
	@Mock
	private UserProfileService userProfileService;
	@Mock
	private UserProfile callContextUser;
	@Mock
	private ClientAttributeService clientAttributeService;
	@Mock
	private BuildProperties buildProperties;

	@BeforeEach
	void init() {
		ReflectionTestUtils.setField(service, "remoteService", Optional.of(remoteService));
	}

	@Nested
	class TestCreateBescheid {

		private final Vorgang vorgang = VorgangTestFactory.create();
		private final Bescheid bescheid = BescheidTestFactory.createBuilder().vorgangId(null).build();

		@BeforeEach
		void initMocks() {
			when(vorgangService.getById(any())).thenReturn(vorgang);
			when(remoteService.create(any(), any())).thenReturn(bescheid);
		}

		@Test
		void shouldLoadVorgang() {
			service.createBescheid(BescheidRequestTestFactory.create());

			verify(vorgangService).getById(VorgangTestFactory.ID);
		}

		@Test
		void shouldCreateBescheid() {
			BescheidRequest request = BescheidRequestTestFactory.create();

			service.createBescheid(request);

			verify(remoteService).create(request, vorgang);
		}

		@Test
		void shouldReturnCreatedBescheid() {
			var result = service.createBescheid(BescheidRequestTestFactory.create());

			assertThat(result).usingRecursiveComparison().isEqualTo(BescheidTestFactory.create());
		}
	}

	@Nested
	class TestSendBescheidManually {

		private final AttachedItem bescheidItem = AttachedItemTestFactory.createBescheid();

		@BeforeEach
		void init() {
			doNothing().when(service).validateBescheidSendManually(any(), anyLong());
		}

		@Test
		void shouldCallValidateBescheid() {
			sendBescheid();

			verify(service).validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldCallSendBescheid() {
			sendBescheid();

			verify(service).sendBescheid(bescheidItem);
		}

		private void sendBescheid() {
			service.sendBescheidManually(bescheidItem, AttachedItemTestFactory.VERSION);
		}
	}

	@Nested
	class TestValidateBescheidSendManually {

		@Test
		void shouldCallValidateBescheid() {
			var bescheidItem = AttachedItemTestFactory.createBescheid();

			service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION);

			verify(service).validateBescheid(bescheidItem, AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldValidate() {
			doNothing().when(service).validateBescheid(any(), anyLong());
			var bescheidItem = AttachedItemTestFactory.createBescheid();

			assertDoesNotThrow(() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION));
		}

		@DisplayName("should decline when")
		@ParameterizedTest(name = "sendBy = {0}")
		@EnumSource(value = Bescheid.SendBy.class, names = "MANUAL", mode = EnumSource.Mode.EXCLUDE)
		void shouldDeclineSendByNotManual(Bescheid.SendBy sendBy) {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().itemEntry(Bescheid.FIELD_SEND_BY, sendBy.name()).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("sendBy");
		}
	}

	@Nested
	class TestSendBescheidPostfachMail {

		private AttachedItem bescheidItem;
		@Mock
		private AttachedItem bescheidSendItem;
		@Captor
		private ArgumentCaptor<VorgangId> vorgangIdCaptor;

		@BeforeEach
		void init() {
			var item = AttachedItemTestFactory.createBescheidItem();
			item.put(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name());
			bescheidItem = AttachedItemTestFactory.createBescheidBuilder().item(item).build();

			doNothing().when(service).validateBescheidSendPostfach(any(), anyLong());
			lenient().when(callContextUser.getId()).thenReturn(UserId.from("user-id"));
			lenient().when(userProfileService.getUserProfile()).thenReturn(callContextUser);
		}

		@Test
		void shouldCallValidateBescheid() {
			doReturn(BescheidTestFactory.create()).when(service).buildBescheid(any(), any());
			when(vorgangService.getById(any())).thenReturn(VorgangTestFactory.create());

			sendBescheid();

			verify(service).validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldCallVorgangService() {
			doReturn(BescheidTestFactory.create()).when(service).buildBescheid(any(), any());
			when(vorgangService.getById(any())).thenReturn(VorgangTestFactory.create());

			sendBescheid();

			verify(vorgangService).getById(vorgangIdCaptor.capture());
			assertThat(vorgangIdCaptor.getValue()).hasToString(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallBuildBescheid() {
			doReturn(BescheidTestFactory.create()).when(service).buildBescheid(any(), any());
			var serviceKonto = ServiceKontoTestFactory.create();
			var vorgang = VorgangTestFactory.createBuilder().serviceKonto(serviceKonto).build();
			doReturn(vorgang).when(vorgangService).getById(any());

			sendBescheid();

			verify(service).buildBescheid(bescheidItem, serviceKonto);
		}

		@Test
		void shouldCallNachrichtService() {
			var bescheid = BescheidTestFactory.create();
			doReturn(bescheid).when(service).buildBescheid(any(), any());
			when(vorgangService.getById(any())).thenReturn(VorgangTestFactory.create());

			sendBescheid();

			verify(nachrichtService).sendNachricht(bescheid);
		}

		@Test
		void shouldCallSendBescheid() {
			doReturn(BescheidTestFactory.create()).when(service).buildBescheid(any(), any());
			when(vorgangService.getById(any())).thenReturn(VorgangTestFactory.create());

			sendBescheid();

			verify(service).sendBescheid(bescheidItem);
		}

		private void sendBescheid() {
			service.sendBescheidPostfachMail(bescheidItem, AttachedItemTestFactory.VERSION);
		}
	}

	@Nested
	class TestValidateBescheidSendPostfach {

		@Test
		void shouldCallValidateBescheid() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, AttachedItemTestFactory.NACHRICHT_SUBJECT)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, AttachedItemTestFactory.NACHRICHT_TEXT).build();

			service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION);

			verify(service).validateBescheid(bescheidItem, AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldValidate() {
			doNothing().when(service).validateBescheid(any(), anyLong());
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, AttachedItemTestFactory.NACHRICHT_SUBJECT)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, AttachedItemTestFactory.NACHRICHT_TEXT).build();

			assertDoesNotThrow(() -> service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION));
		}

		@DisplayName("should decline when")
		@ParameterizedTest(name = "sendBy = {0}")
		@EnumSource(value = Bescheid.SendBy.class, names = "NACHRICHT", mode = EnumSource.Mode.EXCLUDE)
		void shouldDeclineSendByNotPostfach(Bescheid.SendBy sendBy) {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, sendBy.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, AttachedItemTestFactory.NACHRICHT_SUBJECT)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, AttachedItemTestFactory.NACHRICHT_TEXT).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("sendBy");
		}

		@Test
		void shouldDeclineWhenNoSubject() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, null)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, AttachedItemTestFactory.NACHRICHT_TEXT).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("nachricht subject");
		}

		@Test
		void shouldDeclineWhenNoText() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, AttachedItemTestFactory.NACHRICHT_SUBJECT)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, null).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("nachricht text");
		}
	}

	@Nested
	class TestValidateBescheid {

		@Test
		void shouldValidateBescheid() {
			var bescheidItem = AttachedItemTestFactory.createBescheid();

			assertDoesNotThrow(() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION));
		}

		@DisplayName("should decline when")
		@ParameterizedTest(name = "status = {0}")
		@EnumSource(value = Bescheid.Status.class, names = "DRAFT", mode = EnumSource.Mode.EXCLUDE)
		void shouldDeclineNotDraft(Bescheid.Status status) {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().itemEntry(Bescheid.FIELD_STATUS, status.name()).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("has status");
		}

		@Test
		void shouldDeclineWhenNoDocument() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().itemEntry(Bescheid.FIELD_BESCHEID_DOCUMENT, null).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("document");
		}

		@Test
		void shouldDeclineWhenVersionMismatch() {
			var bescheidItem = AttachedItemTestFactory.createBescheid();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION + 1)).getMessage();
			assertThat(message).contains("version");
		}
	}

	@Nested
	class TestBuildBescheid {

		private static final AttachedItem BESCHEID_ITEM = AttachedItemTestFactory.createBescheid();
		private static final Vorgang.ServiceKonto SERVICE_KONTO = ServiceKontoTestFactory.create();
		private static final String CREATED_BY = "user-id";

		@BeforeEach
		void init() {
			when(callContextUser.getId()).thenReturn(UserId.from(CREATED_BY));
			when(userProfileService.getUserProfile()).thenReturn(callContextUser);
			doReturn(FileId.from(DocumentTestFactory.DOCUMENT_FILE)).when(service).getBescheidFileId(any());
		}

		@Test
		void shouldSetVorgangId() {
			var result = buildBescheid();

			assertThat(result.getVorgangId()).hasToString(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallGetBewilligt() {
			buildBescheid();

			verify(service).getBewilligt(BESCHEID_ITEM);
		}

		@Test
		void shouldSetGenehmigt() {
			doReturn(true).when(service).getBewilligt(any());

			var result = buildBescheid();

			assertThat(result.isGenehmigt()).isTrue();
		}

		@Test
		void shouldCallGetBescheidFileId() {
			buildBescheid();

			verify(service).getBescheidFileId(BESCHEID_ITEM);
		}

		@Test
		void shouldSetBescheidFileId() {
			var result = buildBescheid();

			assertThat(result.getBescheidFileId()).hasToString(DocumentTestFactory.DOCUMENT_FILE);
		}

		@Test
		void shouldSetNachrichtSubject() {
			var result = buildBescheid();

			assertThat(result.getNachrichtSubject()).contains(AttachedItemTestFactory.NACHRICHT_SUBJECT);
		}

		@Test
		void shouldSetNachrichtText() {
			var result = buildBescheid();

			assertThat(result.getNachrichtText()).contains(AttachedItemTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldCallCurrentUserService() {
			buildBescheid();

			verify(userProfileService).getUserProfile();
		}

		@Test
		void shouldSetCreatedBy() {
			var result = buildBescheid();

			assertThat(result.getCreatedBy()).hasToString(CREATED_BY);
		}

		@Test
		void shouldSetServiceKonto() {
			var result = buildBescheid();

			assertThat(result.getServiceKonto()).usingRecursiveComparison().isEqualTo(SERVICE_KONTO);
		}

		private Bescheid buildBescheid() {
			return service.buildBescheid(BESCHEID_ITEM, SERVICE_KONTO);
		}
	}

	@Nested
	class TestGetBescheidFileId {

		@BeforeEach
		void init() {
			when(attachedItemService.getItem(anyString())).thenReturn(AttachedItemTestFactory.createDocument());
		}

		@Test
		void shouldCallGetItem() {
			service.getBescheidFileId(AttachedItemTestFactory.createBescheid());

			verify(attachedItemService).getItem(AttachedItemTestFactory.BESCHEID_DOCUMENT);
		}

		@Test
		void shouldReturnBescheidFileId() {
			var result = service.getBescheidFileId(AttachedItemTestFactory.createBescheid());

			assertThat(result).hasToString(DocumentTestFactory.DOCUMENT_FILE);
		}
	}

	@Nested
	class TestGetNachrichtSubject {

		@Test
		void shouldReturnSubject() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().item(Map.of(Bescheid.FIELD_NACHRICHT_SUBJECT,
					AttachedItemTestFactory.NACHRICHT_SUBJECT)).build();

			var result = service.getNachrichtSubject(bescheidItem);

			assertThat(result).isEqualTo(AttachedItemTestFactory.NACHRICHT_SUBJECT);
		}
	}

	@Nested
	class TestGetNachrichtText {

		@Test
		void shouldReturnText() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().item(Map.of(Bescheid.FIELD_NACHRICHT_TEXT,
					AttachedItemTestFactory.NACHRICHT_TEXT)).build();

			var result = service.getNachrichtText(bescheidItem);

			assertThat(result).isEqualTo(AttachedItemTestFactory.NACHRICHT_TEXT);
		}
	}

	@Nested
	class TestSendBescheid {

		private final AttachedItem bescheidItem = AttachedItemTestFactory.createBescheid();
		@Mock
		private AttachedItem bescheidSendItem;

		@Test
		void shouldCallSetBescheidStatusSent() {
			sendBescheid();

			verify(attachedItemService).setBescheidStatus(AttachedItemTestFactory.ID, AttachedItemTestFactory.VERSION, Bescheid.Status.SENT);
		}

		@Test
		void shouldCallBescheiden() {
			sendBescheid();

			verify(vorgangService).bescheiden(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallSetBescheidStatusDraft() {
			doThrow(new TechnicalException("error")).when(vorgangService).bescheiden(anyString());

			assertThrows(TechnicalException.class, this::sendBescheid);

			verify(attachedItemService).setBescheidStatus(AttachedItemTestFactory.ID, AttachedItemTestFactory.VERSION, Bescheid.Status.SENT);
		}

		@Test
		void shouldCallGetBewilligt() {
			sendBescheid();

			verify(service).getBewilligt(bescheidItem);
		}

		@Test
		void shouldCallClientAttributeService() {
			doReturn(true).when(service).getBewilligt(any());

			sendBescheid();

			verify(clientAttributeService).setAntragResult(CommandTestFactory.VORGANG_ID, true);
		}

		private void sendBescheid() {
			service.sendBescheid(bescheidItem);
		}
	}

	@Nested
	class TestGetBewilligt {

		@Test
		void shouldReturnBewilligt() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().itemEntry(Bescheid.FIELD_BEWILLIGT, true).build();

			var result = service.getBewilligt(bescheidItem);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnDefault() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().clearItem().build();

			var result = service.getBewilligt(bescheidItem);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestGetConfig {

		@Test
		void shouldCallGetVersion() {
			service.getConfig();

			verify(buildProperties).getVersion();
		}

		@Test
		void shouldSetBescheidManagerVersion() {
			when(buildProperties.getVersion()).thenReturn(BescheidManagerConfigTestFactory.VERSION);

			var result = service.getConfig();

			assertThat(result.getVersion()).isEqualTo(BescheidManagerConfigTestFactory.VERSION);
		}

		@Test
		void shouldCallGetJavaVersion() {
			service.getConfig();

			verify(service).getJavaVersion();
		}

		@Test
		void shouldSetJavaVersion() {
			when(service.getJavaVersion()).thenReturn(BescheidManagerConfigTestFactory.JAVA_VERSION);

			var result = service.getConfig();

			assertThat(result.getJavaVersion()).isEqualTo(BescheidManagerConfigTestFactory.JAVA_VERSION);
		}

		@Test
		void shouldCallBuildFeatures() {
			service.getConfig();

			verify(service).buildFeatures();
		}

		@Test
		void shouldSetFeatures() {
			var features = BescheidManagerConfig.Features.builder().build();
			when(service.buildFeatures()).thenReturn(features);

			var result = service.getConfig();

			assertThat(result.getFeatures()).isSameAs(features);
		}

	}

	@Nested
	class TestBuildFeatures {

		@Test
		void shouldSetCannotCreateBescheidDocument() {
			ReflectionTestUtils.setField(service, "remoteService", Optional.empty());

			var result = service.buildFeatures();

			assertThat(result.isCanCreateBescheidDocument()).isFalse();
		}

		@Test
		void shouldSetCanCreateBescheidDocument() {
			var result = service.buildFeatures();

			assertThat(result.isCanCreateBescheidDocument()).isTrue();
		}

	}
}
