/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid.attacheditem;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.bescheid.Bescheid;
import de.ozgcloud.bescheid.GrpcBescheid;
import de.ozgcloud.document.DocumentService;
import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;

class AttachedItemMapperTest {

	@Spy
	@InjectMocks
	private AttachedItemMapper mapper = Mappers.getMapper(AttachedItemMapper.class);

	@Mock
	private GrpcObjectMapper grpcObjectMapper;

	@Nested
	class TestMapFromVorgangAttachedItem {

		@BeforeEach
		void init() {
			when(grpcObjectMapper.mapFromGrpc(any())).thenReturn(AttachedItemTestFactory.createDocumentItem());
		}

		@Test
		void shouldMapFromVorgangAttachedItem() {
			var grpcVorgangAttachedItem = GrpcVorgangAttachedItemTestFactory.createBuilder().setItemName(DocumentService.DOCUMENT_ITEM_NAME)
					.setVersion(AttachedItemTestFactory.VERSION).build();

			var result = mapper.mapFromVorgangAttachedItem(grpcVorgangAttachedItem);

			assertThat(result).usingRecursiveComparison().isEqualTo(AttachedItemTestFactory.createDocument());
		}
	}

	@Nested
	class TestToBescheid {

		@Test
		void shouldSetBeschiedenAm() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getBeschiedenAm()).isEqualTo(AttachedItemTestFactory.BESCHEIDEN_AM);
		}

		@Test
		void shouldSetBewilligt() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getBewilligt()).isTrue();
		}

		@Test
		void shouldSetBescheidDocument() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getBescheidDocument()).isEqualTo(AttachedItemTestFactory.BESCHEID_DOCUMENT);
		}

		@Test
		void shouldSetSendBy() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getSendBy()).isEqualTo(AttachedItemTestFactory.SEND_BY.name());
		}

		@Test
		void shouldSetNachrichtText() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getNachrichtText()).isEqualTo(AttachedItemTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldSetNachrichtSubject() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getNachrichtSubject()).isEqualTo(AttachedItemTestFactory.NACHRICHT_SUBJECT);
		}

		@Test
		void shouldCallToAttachments() {
			var expectedAttachments = List.of(AttachedItemTestFactory.ATTACHMENT);
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.item(Map.of(Bescheid.FIELD_ATTACHMENTS, expectedAttachments)).build();

			mapper.toBescheid(bescheidItem);

			verify(mapper).toAttachments(expectedAttachments);
		}

		@Test
		void shouldNotCallToAttachments() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().clearItem().build();

			mapper.toBescheid(bescheidItem);

			verify(mapper, never()).toAttachments(any());
		}

		@Test
		void shouldSetAttachments() {
			doReturn(List.of(AttachedItemTestFactory.ATTACHMENT)).when(mapper).toAttachments(any());

			var bescheid = mapBescheid();

			assertThat(bescheid.getAttachmentsList()).containsExactly(AttachedItemTestFactory.ATTACHMENT);
		}

		@Test
		void shouldSetId() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getId()).isEqualTo(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldSetVersion() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getVersion()).isEqualTo(AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldSetStatus() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getStatus()).isEqualTo(Bescheid.Status.DRAFT.name());
		}

		private GrpcBescheid mapBescheid() {
			return mapper.toBescheid(AttachedItemTestFactory.createBescheid());
		}
	}

	@Nested
	class TestToAttachments {

		@Test
		void shouldMapCollection() {
			var result = mapper.toAttachments(List.of(AttachedItemTestFactory.ATTACHMENT));

			assertThat(result).containsExactly(AttachedItemTestFactory.ATTACHMENT);
		}

		@Test
		void shouldMapSingleValue() {
			var result = mapper.toAttachments(AttachedItemTestFactory.ATTACHMENT);

			assertThat(result).containsExactly(AttachedItemTestFactory.ATTACHMENT);
		}

		@DisplayName("should return null")
		@ParameterizedTest(name = "when attachmentsObject is \"{0}\"")
		@NullAndEmptySource
		void shouldReturnNull(String attachmentsObject) {
			var result = mapper.toAttachments(attachmentsObject);

			assertThat(result).isNull();
		}
	}
}