package de.ozgcloud.bescheid.nachricht;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.apilib.common.datatypes.GenericId;
import de.ozgcloud.apilib.user.OzgCloudUserId;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.bescheid.Bescheid;
import de.ozgcloud.bescheid.BescheidTestFactory;
import de.ozgcloud.bescheid.vorgang.PostfachAddressTestFactory;
import de.ozgcloud.bescheid.vorgang.Vorgang;
import de.ozgcloud.bescheid.vorgang.VorgangTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;

class NachrichtServiceTest {

	@Spy
	@InjectMocks
	private NachrichtService service;

	@Mock
	private NachrichtRemoteService remoteService;
	@Mock
	private OzgCloudCommandService commandService;
	@Mock
	private CommandMapper commandMapper;

	@Nested
	class TestCreateNachrichtDraft {

		private final Nachricht nachricht = NachrichtTestFactory.create();

		@Test
		void shouldCallRemoteService() {
			doReturn(Optional.of(nachricht)).when(service).buildNachricht(any());

			service.createNachrichtDraft(BescheidTestFactory.create());

			verify(remoteService).saveDraft(nachricht);
		}

		@Nested
		class BuildNachricht {

			@BeforeEach
			void init() {
				doReturn(NachrichtTestFactory.MAIL_BODY).when(service).buildMessage(any());
			}

			@Test
			void shouldFillMailBody() {
				var nachricht = service.buildNachricht(BescheidTestFactory.create()).get();

				assertThat(nachricht.getMailBody()).isEqualTo(NachrichtTestFactory.MAIL_BODY);
			}

			@Test
			void shouldSetSubject() {
				var nachricht = service.buildNachricht(BescheidTestFactory.create()).get();

				assertThat(nachricht.getSubject()).isEqualTo(BescheidTestFactory.NACHRICHT_SUBJECT);
			}

			@Test
			void shouldSetDefaultSubject() {
				var bescheid = BescheidTestFactory.createBuilder().nachrichtSubject(Optional.empty()).build();

				var nachricht = service.buildNachricht(bescheid).get();

				assertThat(nachricht.getSubject()).isEqualTo(NachrichtService.SUBJECT);
			}

			@Test
			void shouldSetUser() {
				var nachricht = service.buildNachricht(BescheidTestFactory.create()).get();

				assertThat(nachricht.getCreatedBy()).isEqualTo(BescheidTestFactory.CREATED_BY);
			}

			@Test
			void shouldSetVorgangId() {
				var nachricht = service.buildNachricht(BescheidTestFactory.create()).get();

				assertThat(nachricht.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
			}
		}
	}

	@Nested
	class TestBuildMessage {
		@Test
		void shouldUseTextFromBescheid() {
			var message = service.buildMessage(BescheidTestFactory.create());

			assertThat(message).isEqualTo(BescheidTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldUseDefaultTemplate() {
			doReturn("FROM_TEMPLATE").when(service).fillTemplate(any(), any());

			var message = service.buildMessage(BescheidTestFactory.createBuilder().nachrichtText(Optional.empty()).build());

			assertThat(message).isEqualTo("FROM_TEMPLATE");
		}
	}

	@Nested
	class TestSendNachricht {

		@Mock
		private OzgCloudCommand ozgCloudCommand;

		@BeforeEach
		void init() {
			doReturn(ozgCloudCommand).when(service).buildSendBescheidCommand(any());
		}

		@Test
		void shouldCallBuildBescheidCommand() {
			var bescheid = BescheidTestFactory.create();

			service.sendNachricht(bescheid);

			verify(service).buildSendBescheidCommand(bescheid);
		}

		@Test
		void shouldCallCommandService() {
			service.sendNachricht(BescheidTestFactory.create());

			verify(commandService).createAndWaitUntilDone(ozgCloudCommand);
		}

	}

	@Nested
	class TestBuildSendBescheidCommand {

		private static final Map<String, Object> bodyObject = Map.of("key", "value");
		private final Bescheid bescheid = BescheidTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(bodyObject).when(service).buildSendNachrichtCommandBody(any());
		}

		@Test
		void shouldCallVorgangIdMapper() {
			buildCommand();

			verify(commandMapper).toOzgCloudVorgangId(VorgangTestFactory.ID.toString());
		}

		@Test
		void shouldSetVorgangId() {
			var vorgangId = OzgCloudVorgangId.from(VorgangTestFactory.ID.toString());
			when(commandMapper.toOzgCloudVorgangId(anyString())).thenReturn(vorgangId);

			var result = buildCommand();

			assertThat(result.getVorgangId()).isEqualTo(vorgangId);
		}

		@Test
		void shouldCallMapRelationId() {
			buildCommand();

			verify(commandMapper).mapRelationId(VorgangTestFactory.ID.toString());
		}

		@Test
		void shouldSetRelationId() {
			var relationId = GenericId.from("relation-id");
			when(commandMapper.mapRelationId(anyString())).thenReturn(relationId);

			var result = buildCommand();

			assertThat(result.getRelationId()).isEqualTo(relationId);
		}

		@Test
		void shouldSetOrder() {
			var result = buildCommand();

			assertThat(result.getOrder()).isEqualTo(NachrichtService.SEND_POSTFACH_NACHRICHT_ORDER);
		}

		@Test
		void shouldCallCreatedBy() {
			buildCommand();

			verify(commandMapper).toOzgCloudUserId(BescheidTestFactory.CREATED_BY.toString());
		}

		@Test
		void shouldSetCreatedBy() {
			var userId = OzgCloudUserId.from(BescheidTestFactory.CREATED_BY.toString());
			when(commandMapper.toOzgCloudUserId(anyString())).thenReturn(userId);

			var result = buildCommand();

			assertThat(result.getCreatedBy()).isEqualTo(userId);
		}

		@Test
		void shouldCallBuildSendNachrichtCommandBody() {
			buildCommand();

			verify(service).buildSendNachrichtCommandBody(bescheid);
		}

		@Test
		void shouldSetBodyObject() {
			var result = buildCommand();

			assertThat(result.getBodyObject()).isEqualTo(bodyObject);
		}

		private OzgCloudCommand buildCommand() {
			return service.buildSendBescheidCommand(bescheid);
		}
	}

	@Nested
	class TestBuildSendNachrichtCommandBody {

		private static final String POSTFACH_ID = "postfach-id";

		private final Bescheid bescheid = BescheidTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(POSTFACH_ID).when(service).getPostfachId(any());
		}

		@Test
		void shouldSetPostfachId() {
			var result = service.buildSendNachrichtCommandBody(bescheid);

			assertThat(result).containsEntry(NachrichtService.FIELD_POSTFACH_ID, POSTFACH_ID);
		}

		@Test
		void shouldSetReplyOption() {
			var result = service.buildSendNachrichtCommandBody(bescheid);

			assertThat(result).containsEntry(NachrichtService.FIELD_REPLY_OPTION, NachrichtService.REPLY_OPTION);
		}

		@Test
		void shouldSetSubject() {
			var result = service.buildSendNachrichtCommandBody(bescheid);

			assertThat(result).containsEntry(NachrichtService.FIELD_SUBJECT, BescheidTestFactory.NACHRICHT_SUBJECT);
		}

		@Test
		void shouldSetMailBody() {
			var result = service.buildSendNachrichtCommandBody(bescheid);

			assertThat(result).containsEntry(NachrichtService.FIELD_MAIL_BODY, BescheidTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldSetAttachments() {
			var result = service.buildSendNachrichtCommandBody(bescheid);

			assertThat(result).containsEntry(NachrichtService.FIELD_ATTACHMENTS, BescheidTestFactory.BESCHEID_FILE_ID);
		}
	}

	@Nested
	class TestGetPostfachId {

		private static final Vorgang.PostfachAddress POSTFACH_ADDRESS = PostfachAddressTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(Optional.of(POSTFACH_ADDRESS)).when(service).getAddress(any());
		}

		@Test
		void shouldCallGetAddress() {
			var bescheid = BescheidTestFactory.create();

			service.getPostfachId(bescheid);

			verify(service).getAddress(bescheid);
		}

		@Test
		void shouldReturnPostfachId() {
			var bescheid = BescheidTestFactory.create();

			var result = service.getPostfachId(bescheid);

			assertThat(result).isEqualTo(PostfachAddressTestFactory.POSTFACH_ID);
		}

	}

	@Nested
	class TestNoPostfachId {

		@Test
		void shouldThrowExceptionIfNoPostfachAddress() {
			doReturn(Optional.empty()).when(service).getAddress(any());

			assertThrows(TechnicalException.class, this::getPostfachId);
		}

		@Test
		void shouldThrowExceptionIfNoPostfachId() {
			var postfachAddress = PostfachAddressTestFactory.createBuilder().identifier(Map.of()).build();
			doReturn(Optional.of(postfachAddress)).when(service).getAddress(any());

			assertThrows(TechnicalException.class, this::getPostfachId);
		}

		private void getPostfachId() {
			service.getPostfachId(BescheidTestFactory.create());
		}
	}
}
