package de.ozgcloud.bescheid.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.apilib.common.datatypes.GenericId;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;

class VorgangServiceTest {

	@Spy
	@InjectMocks
	private VorgangService service;

	@Mock
	private VorgangRemoteService remoteService;
	@Mock
	private OzgCloudCommandService commandService;
	@Mock
	private CommandMapper commandMapper;

	@Nested
	class TestGetById {

		private final Vorgang vorgang = VorgangTestFactory.create();

		@BeforeEach
		void init() {
			when(remoteService.getById(any())).thenReturn(vorgang);
		}

		@Test
		void shouldCallRemoteService() {
			service.getById(VorgangTestFactory.ID);

			verify(remoteService).getById(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnResult() {
			var result = service.getById(VorgangTestFactory.ID);

			assertThat(result).isEqualTo(vorgang);
		}
	}

	@Nested
	class TestBescheiden {

		private final static Vorgang VORGANG = VorgangTestFactory.create();

		@Mock
		private OzgCloudCommand ozgCloudCommand;

		@BeforeEach
		void init() {
			doReturn(VORGANG).when(service).getById(any(VorgangId.class));
		}

		@Test
		void shouldCallGetBeyId() {
			service.bescheiden(VorgangTestFactory.ID.toString());

			verify(service).getById(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallBuildBescheidenCommand() {
			service.bescheiden(VorgangTestFactory.ID.toString());

			verify(service).buildBescheidenCommand(VORGANG);
		}

		@Test
		void shouldCallCommandService() {
			doReturn(ozgCloudCommand).when(service).buildBescheidenCommand(any());

			service.bescheiden(VorgangTestFactory.ID.toString());

			verify(commandService).createAndWaitUntilDone(ozgCloudCommand);
		}
	}

	@Nested
	class TestBuildBescheidenCommand {

		private final static Vorgang VORGANG = VorgangTestFactory.create();

		@Test
		void shouldCallCommandMapperToOzgCloudVorgangId() {
			service.buildBescheidenCommand(VORGANG);

			verify(commandMapper).toOzgCloudVorgangId(VorgangTestFactory.ID.toString());
		}

		@Test
		void shouldSetVorgangId() {
			var expectedId = OzgCloudVorgangId.from(VorgangTestFactory.ID.toString());
			when(commandMapper.toOzgCloudVorgangId(any())).thenReturn(expectedId);

			var result = service.buildBescheidenCommand(VORGANG);

			assertThat(result.getVorgangId()).isEqualTo(expectedId);
		}

		@Test
		void shouldCallCommandMapperMapRelationId() {
			service.buildBescheidenCommand(VORGANG);

			verify(commandMapper).mapRelationId(VorgangTestFactory.ID.toString());
		}

		@Test
		void shouldSetRelationId() {
			var expectedId = GenericId.from(VorgangTestFactory.ID.toString());
			when(commandMapper.mapRelationId(any())).thenReturn(expectedId);

			var result = service.buildBescheidenCommand(VORGANG);

			assertThat(result.getRelationId()).isEqualTo(expectedId);
		}

		@Test
		void shouldSetRelationVersion() {
			var expectedVersion = VORGANG.getVersion();

			var result = service.buildBescheidenCommand(VORGANG);

			assertThat(result.getRelationVersion()).isEqualTo(expectedVersion);
		}

		@Test
		void shouldSetOrder() {
			var expectedOrder = VorgangService.VORGANG_BESCHEIDEN;

			var result = service.buildBescheidenCommand(VORGANG);

			assertThat(result.getOrder()).isEqualTo(expectedOrder);
		}
	}
}
