/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import static de.ozgcloud.vorgang.common.db.CriteriaUtil.*;
import static org.springframework.data.mongodb.core.query.Criteria.*;
import static org.springframework.data.mongodb.core.query.Query.*;

import java.util.HashMap;
import java.util.Map;

import jakarta.annotation.PostConstruct;

import org.bson.Document;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;
import org.springframework.stereotype.Repository;

import de.ozgcloud.vorgang.common.db.CollisionVerifier;

@Repository
public class VorgangAttachedItemCustomRepositoryImpl implements VorgangAttachedItemCustomRepository {

	@Autowired
	private MongoOperations mongoOperations;

	private CollisionVerifier collisionVerifier;

	@PostConstruct
	void init() {
		collisionVerifier = createCollisionVerifier();
	}

	CollisionVerifier createCollisionVerifier() {
		return new CollisionVerifier(this::doVorgangAttachedItemExist);
	}

	boolean doVorgangAttachedItemExist(String itemId) {
		return mongoOperations.exists(query(isId(itemId)), VorgangAttachedItem.COLLECTION_NAME);
	}

	@Override
	public void patch(String itemId, long version, Map<String, Object> propertyMap) {
		mongoOperations.updateFirst(
				query(byIdAndVersion(itemId, version)),
				createUpdateForItemAndVersion(propertyMap),
				VorgangAttachedItem.class);

	}

	@Override
	public void forcePatch(String itemId, Map<String, Object> propertiesMap) {
		mongoOperations.updateFirst(
				createQueryById(itemId),
				createUpdateForItemAndVersion(propertiesMap),
				VorgangAttachedItem.class);
	}

	private Query createQueryById(String itemId) {
		return query(new Criteria().andOperator(byId(itemId), byNotDeleted()));
	}

	private Criteria byIdAndVersion(String itemId, long version) {
		return new Criteria().andOperator(byId(itemId), byVersion(version), byNotDeleted());
	}

	private Criteria byId(String itemId) {
		return where(VorgangAttachedItem.FIELDNAME_ID).is(itemId);
	}

	private Criteria byVersion(long version) {
		return where(VorgangAttachedItem.FIELDNAME_VERSION).is(version);
	}

	private Criteria byNotDeleted() {
		return where(VorgangAttachedItem.FIELDNAME_IS_DELETED).is(false);
	}

	private Update createUpdateForItemAndVersion(Map<String, Object> propertyMap) {
		return Update.fromDocument(buildItemPatchDocument(propertyMap)).inc(VorgangAttachedItem.FIELDNAME_VERSION, 1);
	}

	private Document buildItemPatchDocument(Map<String, Object> propertiesMap) {
		Map<String, Object> resultMap = new HashMap<>(propertiesMap.size());
		propertiesMap.forEach((k, v) -> resultMap.put("%s.%s".formatted(VorgangAttachedItem.FIELDNAME_ITEM, k), v));
		return new Document(resultMap);
	}

	@Override
	public void delete(String itemId, long version) {
		var deleteResult = mongoOperations.remove(query(whereIdAndVersion(itemId, version)), VorgangAttachedItem.COLLECTION_NAME);
		collisionVerifier.verify(deleteResult, itemId);
	}

	@Override
	public void updateDeleted(boolean deleted, String itemId, long version) {
		var updateResult = mongoOperations.updateFirst(query(whereIdAndVersion(itemId, version)),
				Update.update(VorgangAttachedItem.FIELDNAME_IS_DELETED, deleted), VorgangAttachedItem.COLLECTION_NAME);
		collisionVerifier.verify(updateResult, itemId);
	}

}