/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import io.grpc.stub.StreamObserver;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.extern.log4j.Log4j2;

@Log4j2
class UploadStreamObserver implements StreamObserver<GrpcUploadBinaryFileRequest> {
	private final FileService service;
	private final PolicyService policyService;
	private final StreamObserver<GrpcUploadBinaryFileResponse> responseObserver;
	@Getter(value = AccessLevel.PACKAGE)
	private PipedOutputStream pipedOutput;
	@Getter(value = AccessLevel.PACKAGE)
	private PipedInputStream pipedInput;

	private CompletableFuture<FileId> fileIdFuture;

	public UploadStreamObserver(StreamObserver<GrpcUploadBinaryFileResponse> responseObserver, FileService service, PolicyService policyService) {
		this.responseObserver = responseObserver;
		this.service = service;
		this.policyService = policyService;
	}

	@Override
	public void onNext(GrpcUploadBinaryFileRequest fileUploadRequest) {
		if (fileUploadRequest.hasMetadata()) {
			policyService.checkPermission(Set.of(fileUploadRequest.getMetadata().getVorgangId()));
			initPipedStreams();
			fileIdFuture = storeFileMetaData(fileUploadRequest, pipedInput, pipedOutput);
		} else {
			storeFileContent(fileUploadRequest, pipedOutput);
		}
	}

	void initPipedStreams() {
		pipedOutput = new PipedOutputStream();
		pipedInput = new PipedInputStream(FileService.CHUNK_SIZE);
	}

	void storeFileContent(GrpcUploadBinaryFileRequest fileUploadRequest, PipedOutputStream pipedOutput) {
		try {
			pipedOutput.write(fileUploadRequest.getFileContent().toByteArray());
		} catch (IOException e) {
			LOG.error("Fehler beim Speichern des Dateiinhalts.", e);
		}
	}

	CompletableFuture<FileId> storeFileMetaData(GrpcUploadBinaryFileRequest fileUploadRequest, PipedInputStream pipedInput,
			PipedOutputStream pipedOutput) {
		connectInputStreamToOutputStream(pipedInput, pipedOutput);

		return storeMetaData(fileUploadRequest);
	}

	void connectInputStreamToOutputStream(PipedInputStream is, PipedOutputStream os) {
		try {
			is.connect(os);
		} catch (IOException e) {
			LOG.warn("Fehler beim Verbinden der Input/Outputs", e);
		}
	}

	CompletableFuture<FileId> storeMetaData(GrpcUploadBinaryFileRequest fileUploadRequest) {
		var userId = fileUploadRequest.getMetadata().getContext().getUser().getId();

		return service.uploadFileStreamAsync(
				buildUploadFilesReferences(fileUploadRequest),
				buildOzgFile(fileUploadRequest),
				Optional.of(userId), pipedInput);
	}

	UploadedFilesReference buildUploadFilesReferences(GrpcUploadBinaryFileRequest fileUploadRequest) {
		return UploadedFilesReference.builder()
				.vorgangId(fileUploadRequest.getMetadata().getVorgangId())
				.client(GrpcBinaryFileService.CLIENT)
				.name(fileUploadRequest.getMetadata().getField()).build();
	}

	OzgFile buildOzgFile(GrpcUploadBinaryFileRequest fileUploadRequest) {
		return OzgFile.builder()
				.name(fileUploadRequest.getMetadata().getFileName())
				.size(fileUploadRequest.getMetadata().getSize())
				.contentType(fileUploadRequest.getMetadata().getContentType())
				.build();
	}

	@Override
	public void onError(Throwable t) {
		LOG.error("GrpcLargeFileRequestObserver onError", t);
		throw new TechnicalException("GrpcLargeFileRequestObserver onError has been called", t);
	}

	@Override
	public void onCompleted() {
		try {
			pipedOutput.close();

			responseObserver.onNext(GrpcBinaryFileService.buildResponse(fileIdFuture.get(10, TimeUnit.MINUTES)));

			pipedInput.close();
		} catch (ExecutionException | IOException | TimeoutException e) {
			handleException(e);
		} catch (InterruptedException ie) {
			handleException(ie);
			Thread.currentThread().interrupt();
		}
		responseObserver.onCompleted();

	}

	void handleException(Exception e) {
		responseObserver.onNext(GrpcUploadBinaryFileResponse.newBuilder().build());
		throw new TechnicalException(e.getMessage(), e);
	}
}