/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid;

import static de.ozgcloud.bescheid.BescheidEventListener.*;
import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.gridfs.GridFsTemplate;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.test.annotation.DirtiesContext;

import com.mongodb.client.gridfs.model.GridFSFile;

import de.ozgcloud.apilib.user.OzgCloudUserId;
import de.ozgcloud.apilib.user.OzgCloudUserProfile;
import de.ozgcloud.apilib.user.OzgCloudUserProfileService;
import de.ozgcloud.bescheid.attacheditem.AttachedItemService;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.document.BescheidDocumentCreatedEvent;
import de.ozgcloud.document.Document;
import de.ozgcloud.document.DocumentService;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachRemoteService;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.callcontext.CallContextAuthenticationTokenTestFactory;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.CreateCommandRequest;
import de.ozgcloud.vorgang.servicekonto.PostfachAddressTestFactory;
import de.ozgcloud.vorgang.status.StatusChangedEvent;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@SpringBootTest(classes = VorgangManagerServerApplication.class, properties = {
		"grpc.server.inProcessName=test",
		"grpc.server.port=-1",
		"grpc.client.ozgcloud-command-manager.address=in-process:test",
		"grpc.client.vorgang-manager.address=in-process:test",
		"grpc.client.nachrichten-manager.address=in-process:test",
		"grpc.client.pluto.address=in-process:test",
		"grpc.client.command-manager.address=in-process:test",
		"ozgcloud.feature.bescheid.enable-dummy-document-processor=true",
})
@DataITCase
@DirtiesContext
class BescheidEventListenerITCase {

	@Autowired
	private CommandService commandService;
	@Autowired
	private ApplicationEventPublisher eventPublisher;
	@SpyBean
	private TestEventListener bescheiTestEventListener;

	@Autowired
	private MongoOperations mongoOperations;
	@Autowired
	private GridFsTemplate gridFsTemplate;

	@MockBean
	private PostfachRemoteService postfachRemoteService;
	@MockBean
	private OzgCloudUserProfileService userProfileService;

	@Captor
	private ArgumentCaptor<BescheidDocumentCreatedEvent> bescheidDocumentCreatedEventCaptor;

	@BeforeEach
	void init() {
		mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
		mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
		when(postfachRemoteService.getPostfachType()).thenReturn("dummy");
		when(userProfileService.getById(any())).thenReturn(OzgCloudUserProfile.builder().id(OzgCloudUserId.from("user-id")).build());
	}

	@Nested
	class TestCreateBescheidDocument {

		@Test
		void shouldCreateBescheidDocument() {
			var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var bescheidItemId = mongoOperations.save(createBescheidAttachedItem(vorgangId)).getId();

			eventPublisher.publishEvent(buildCommandCreatedEvent(vorgangId, bescheidItemId));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidDocumentCreated(bescheidDocumentCreatedEventCaptor.capture());
				assertThat(getBescheidDocument()).isNotNull();
			});
		}

		@Test
		void shouldCreateDocumentFile() {
			var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var bescheidItemId = mongoOperations.save(createBescheidAttachedItem(vorgangId)).getId();

			eventPublisher.publishEvent(buildCommandCreatedEvent(vorgangId, bescheidItemId));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidDocumentCreated(bescheidDocumentCreatedEventCaptor.capture());
				assertThat(getBescheidDocumentFile()).isNotNull();
			});
		}

		private VorgangAttachedItem createBescheidAttachedItem(String vorgangId) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(null)
					.version(0L)
					.vorgangId(vorgangId)
					.itemName(AttachedItemService.BESCHEID_ITEM_NAME)
					.item(Map.of(Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name()))
					.build();
		}

		private CommandCreatedEvent buildCommandCreatedEvent(String vorgangId, String bescheidItemId) {
			var command = CommandTestFactory.createBuilder()
					.vorgangId(vorgangId)
					.relationId(bescheidItemId)
					.order(BescheidEventListener.CREATE_BESCHEID_DOCUMENT_ORDER)
					.bodyObject(Map.of(BESCHEID_VOM_BODYKEY, "2024-01-01",
							GENEHMIGT_BODYKEY, true))
					.build();
			return new CommandCreatedEvent(command);
		}

		private GridFSFile getBescheidDocumentFile() {
			var fileId = getBescheidDocument().getItem().get(Document.FIELD_DOCUMENT_FILE).toString();
			return gridFsTemplate.findOne(Query.query(Criteria.where("_id").is(fileId)));
		}

		private VorgangAttachedItem getBescheidDocument() {
			var documentId = bescheidDocumentCreatedEventCaptor.getValue().getCreatedResource();
			return mongoOperations.findById(documentId, VorgangAttachedItem.class);
		}

	}

	@Nested
	class TestSendBescheidManually {

		@Captor
		private ArgumentCaptor<BescheidSentEvent> bescheidSentEventCaptor;
		@Captor
		private ArgumentCaptor<StatusChangedEvent> statusChangedEventCaptor;

		@Test
		void shouldSetBescheidStatusToSent() {
			var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var bescheidItem = mongoOperations.save(createBescheidAttachedItemWithDocument(vorgangId));

			eventPublisher.publishEvent(buildCommandCreatedEvent(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				assertThat(loadBescheid(bescheidItem.getId()).getItem()).containsEntry(Bescheid.FIELD_STATUS, Bescheid.Status.SENT.name());
			});
		}

		@Test
		void shouldSetVorgangStatusToBeschieden() {
			var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var bescheidItem = mongoOperations.save(createBescheidAttachedItemWithDocument(vorgangId));
			var event = buildCommandCreatedEvent(vorgangId, bescheidItem);

			eventPublisher.publishEvent(event);

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidStatusChangedEvent(statusChangedEventCaptor.capture());
				assertThat(loadVorgang(vorgangId).getStatus()).isEqualTo(Vorgang.Status.BESCHIEDEN);
			});
		}

		private VorgangAttachedItem createBescheidAttachedItemWithDocument(String vorgangId) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(null)
					.version(0L)
					.vorgangId(vorgangId)
					.itemName(AttachedItemService.BESCHEID_ITEM_NAME)
					.item(Map.of(
							Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name(),
							Bescheid.FIELD_SEND_BY, Bescheid.SendBy.MANUAL.name(),
							Bescheid.FIELD_BESCHEID_DOCUMENT, "documentId"))
					.build();
		}

		private CommandCreatedEvent buildCommandCreatedEvent(String vorgangId, VorgangAttachedItem bescheidItem) {
			var command = CommandTestFactory.createBuilder()
					.vorgangId(vorgangId)
					.relationId(bescheidItem.getId())
					.relationVersion(bescheidItem.getVersion())
					.order(BescheidEventListener.SEND_BESCHEID_ORDER)
					.bodyObject(Map.of(BESCHEID_VOM_BODYKEY, "2024-01-01",
							GENEHMIGT_BODYKEY, true))
					.build();
			return new CommandCreatedEvent(command);
		}

		private VorgangAttachedItem loadBescheid(String bescheidId) {
			return mongoOperations.findById(bescheidId, VorgangAttachedItem.class);
		}

		private Vorgang loadVorgang(String vorgangId) {
			return mongoOperations.findById(vorgangId, Vorgang.class);
		}
	}

	@Nested
	class TestSendBescheidPostfachMail {

		private static final String NACHRICHT_SUBJECT = "nachrichtSubject";
		private static final String NACHRICHT_TEXT = "nachrichtText";

		@Captor
		private ArgumentCaptor<BescheidSentEvent> bescheidSentEventCaptor;
		@Captor
		private ArgumentCaptor<StatusChangedEvent> statusChangedEventCaptor;
		@Captor
		private ArgumentCaptor<PostfachNachricht> postfachNachrichtCaptor;

		private String vorgangId;
		private VorgangAttachedItem bescheidItem;

		@BeforeEach
		void init() {
			SecurityContextHolder.getContext().setAuthentication(CallContextAuthenticationTokenTestFactory.create());
			vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var documentFileId = mongoOperations.save(createBescheidDocumentAttachedItem(vorgangId)).getId();
			bescheidItem = mongoOperations.save(createBescheidAttachedItemWithDocument(vorgangId, documentFileId));
		}

		@AfterAll
		static void cleanup() {
			SecurityContextHolder.clearContext();
		}

		@Test
		void shouldSuccessfullyCompleteCommand() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				assertThat(loadCommand(bescheidSentEventCaptor.getValue().getSource()).getStatus()).isEqualTo(CommandStatus.FINISHED);
			});
		}

		@Test
		void shouldSetBescheidStatusToSent() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				assertThat(loadBescheid(bescheidItem.getId()).getItem()).containsEntry(Bescheid.FIELD_STATUS, Bescheid.Status.SENT.name());
			});
		}

		@Test
		void shouldSetVorgangStatusToBeschieden() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidStatusChangedEvent(statusChangedEventCaptor.capture());
				assertThat(loadCommand(statusChangedEventCaptor.getValue().getSource()).getStatus()).isEqualTo(CommandStatus.FINISHED);
				assertThat(loadVorgang(vorgangId).getStatus()).isEqualTo(Vorgang.Status.BESCHIEDEN);
			});
		}

		@Test
		void shouldSendPostfachNachricht() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(postfachRemoteService).sendMessage(postfachNachrichtCaptor.capture());
				assertThat(postfachNachrichtCaptor.getValue().getPostfachId()).isEqualTo(
						PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
				assertThat(postfachNachrichtCaptor.getValue().getSubject()).isEqualTo(NACHRICHT_SUBJECT);
				assertThat(postfachNachrichtCaptor.getValue().getMailBody()).isEqualTo(NACHRICHT_TEXT);
			});
		}

		@Test
		void shouldSetClientAttribute() {
			commandService.createCommand(buildCreateCommandRequest(vorgangId, bescheidItem));

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				verify(bescheiTestEventListener).onBescheidSentEvent(bescheidSentEventCaptor.capture());
				assertThat(loadCommand(bescheidSentEventCaptor.getValue().getSource()).getStatus()).isEqualTo(CommandStatus.FINISHED);
				var vorgang = loadVorgang(vorgangId);
				assertThat(vorgang.getClientAttributes()).containsKey(BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);

			});
		}

		private VorgangAttachedItem createBescheidAttachedItemWithDocument(String vorgangId, String bescheidDocumentId) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(null)
					.version(0L)
					.vorgangId(vorgangId)
					.itemName(AttachedItemService.BESCHEID_ITEM_NAME)
					.item(Map.of(
							Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name(),
							Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name(),
							Bescheid.FIELD_BEWILLIGT, true,
							Bescheid.FIELD_BESCHEID_DOCUMENT, bescheidDocumentId,
							Bescheid.FIELD_NACHRICHT_SUBJECT, NACHRICHT_SUBJECT,
							Bescheid.FIELD_NACHRICHT_TEXT, NACHRICHT_TEXT))
					.build();
		}

		private VorgangAttachedItem createBescheidDocumentAttachedItem(String vorgangId) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(null)
					.version(0L)
					.vorgangId(vorgangId)
					.itemName(DocumentService.DOCUMENT_ITEM_NAME)
					.item(Map.of(
							DocumentService.FIELD_DOCUMENT_TYPE, DocumentService.DOCUMENT_TYPE,
							DocumentService.FIELD_DOCUMENT_FILE, "documentId",
							DocumentService.FIELD_NACHRICHT_TEXT, "nachrichtText"))
					.build();
		}

		private CreateCommandRequest buildCreateCommandRequest(String vorgangId, VorgangAttachedItem bescheidItem) {
			return CreateCommandRequest.builder()
					.vorgangId(vorgangId)
					.relationId(bescheidItem.getId())
					.relationVersion(bescheidItem.getVersion())
					.order(SEND_BESCHEID_ORDER)
					.bodyObject(Map.of(BESCHEID_VOM_BODYKEY, "2024-01-01",
							GENEHMIGT_BODYKEY, true))
					.build();
		}

		private Command loadCommand(String commandId) {
			return mongoOperations.findById(commandId, Command.class);
		}

		private VorgangAttachedItem loadBescheid(String bescheidId) {
			return mongoOperations.findById(bescheidId, VorgangAttachedItem.class);
		}

		private Vorgang loadVorgang(String vorgangId) {
			return mongoOperations.findById(vorgangId, Vorgang.class);
		}

	}
}
