/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.concurrent.atomic.AtomicInteger;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.attached_item.GrpcVorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.common.GrpcObject;
import de.ozgcloud.vorgang.common.GrpcProperty;
import de.ozgcloud.vorgang.servicekonto.PostfachAddressTestFactory;
import de.ozgcloud.vorgang.servicekonto.ServiceKontoTestFactory;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangResponse;
import de.ozgcloud.vorgang.vorgang.GrpcVorgangHeader;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemResponse;
import de.ozgcloud.vorgang.vorgangAttachedItem.VorgangAttachedItemServiceGrpc.VorgangAttachedItemServiceBlockingStub;

@SpringBootTest(classes = { VorgangManagerServerApplication.class })
@ITCase
class PersistPostfachNachrichtServiceImplITCase {

	private static final String MESSAGE_ID = "message-id";
	@Autowired
	private PersistPostfachNachrichtServiceImpl service;
	@Autowired
	private VorgangRemoteService vorgangRemoteService;
	@Autowired
	private AttachedItemRemoteService attachedItemRemoteService;

	private AtomicInteger counter = new AtomicInteger(1);
	@Mock
	private VorgangServiceBlockingStub vorgangServiceStub;
	@Mock
	private VorgangAttachedItemServiceBlockingStub vorgangAttachedItemServiceStub;

	@BeforeEach
	void init() {
		when(vorgangServiceStub.withInterceptors(any())).thenReturn(vorgangServiceStub);
		ReflectionTestUtils.setField(vorgangRemoteService, "vorgangServiceStub", vorgangServiceStub);

		when(vorgangAttachedItemServiceStub.withInterceptors(any())).thenReturn(vorgangAttachedItemServiceStub);
		ReflectionTestUtils.setField(attachedItemRemoteService, "vorgangAttachedItemServiceStub", vorgangAttachedItemServiceStub);

		mockFindVorgang();
		mockFindAttachedItems();
	}

	@Test
	void shouldFindRuckfragen() {
		var rueckfragen = service.findRueckfragen(ServiceKontoTestFactory.TYPE, PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);

		assertThat(rueckfragen).hasSize(1).first().extracting(PostfachNachricht::getId).isEqualTo(MESSAGE_ID);
	}

	private void mockFindVorgang() {
		when(vorgangServiceStub.findVorgang(any())).thenAnswer(invocation -> {
			if (counter.getAndDecrement() == 0) {
				return GrpcFindVorgangResponse.newBuilder().build();
			}
			return GrpcFindVorgangResponse.newBuilder().setTotal(1).addVorgang(GrpcVorgangHeader.newBuilder().build()).build();
		});
	}

	private void mockFindAttachedItems() {
		var attachedItemResponse = GrpcFindVorgangAttachedItemResponse.newBuilder()
				.addVorgangAttachedItems(GrpcVorgangAttachedItemTestFactory.create())
				.addVorgangAttachedItems(GrpcVorgangAttachedItemTestFactory.createBuilder().setItem(buildGrpcObject())).build();
		when(vorgangAttachedItemServiceStub.find(any())).thenReturn(attachedItemResponse);
	}

	private GrpcObject buildGrpcObject() {
		return GrpcObject.newBuilder()
				.addProperty(GrpcProperty.newBuilder()
						.setName(PostfachNachricht.FIELD_REPLY_OPTION).addValue(PostfachNachricht.ReplyOption.POSSIBLE.name()).build())
				.addProperty(GrpcProperty.newBuilder()
						.setName(PostfachNachricht.FIELD_CREATED_AT).addValue(PostfachNachrichtTestFactory.CREATED_AT_STR))
				.addProperty(GrpcProperty.newBuilder()
						.setName(PostfachNachricht.FIELD_DIRECTION).addValue(PostfachNachrichtTestFactory.DIRECTION.name()))
				.addProperty(GrpcProperty.newBuilder()
						.setName(PostfachNachricht.FIELD_ID).addValue(MESSAGE_ID).build())
				.build();
	}
}
