package de.ozgcloud.processor;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.test.context.ActiveProfiles;

import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.command.VorgangCreatedEvent;
import de.ozgcloud.notification.postfach.PostfachService;
import de.ozgcloud.processor.processor.ProcessorService;
import de.ozgcloud.processor.vorgang.Vorgang;
import de.ozgcloud.processor.vorgang.VorgangId;
import de.ozgcloud.processor.vorgang.VorgangService;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.callcontext.WithMockCustomUser;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CreateCommandRequest;
import de.ozgcloud.vorgang.command.PersistedCommand;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@SpringBootTest(properties = {
		"ozgcloud.processors.0.name=test",
		"ozgcloud.processors.0.address=http://localhost:8090/testprocessor",
		"ozgcloud.processors.0.forms.0.formEngineName=testFormEngine",
		"ozgcloud.processors.0.forms.0.formId=testForm",
		"grpc.server.port=9091",
		"ozgcloud.command-manager.address=static://127.0.0.1:9091"
}, classes = { VorgangManagerServerApplication.class })
@ActiveProfiles({ "local", "itcase" })
@WithMockCustomUser
class ProcessorITCase {

	@Autowired
	private ApplicationEventPublisher publisher;

	@MockBean
	private VorgangService processorVorgangService;

	@MockBean
	private PostfachService postfachService;
	@MockBean
	private de.ozgcloud.notification.vorgang.VorgangService notificationService;

	@MockBean
	private CommandService commandService;
	@MockBean
	private ProcessorService processorService;

	@Captor
	private ArgumentCaptor<CreateCommandRequest> requestCaptor;

	@BeforeEach
	void init() {
		when(processorVorgangService.getVorgang(any())).thenReturn(buildVorgang());

		when(notificationService.getVorgang(any())).thenReturn(de.ozgcloud.notification.vorgang.Vorgang.builder().build());

		when(processorService.processVorgang(any())).thenThrow(RuntimeException.class);
	}

	@BeforeEach
	void setupCommandService() {
		var commandBuilder = PersistedCommand.builder().id("42").status(CommandStatus.PENDING).createdAt(ZonedDateTime.now());

		when(commandService.createCommand(any())).thenReturn(commandBuilder.build());
		when(commandService.findCommand("42"))
				.thenReturn(Optional.of(commandBuilder.status(CommandStatus.FINISHED).finishedAt(ZonedDateTime.now()).build()));
	}

	@Test
	void triggerProcessorOnNewVorgang() {
		publisher.publishEvent(new VorgangCreatedEvent(VorgangTestFactory.ID));

		await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> {
			verify(processorVorgangService).getVorgang(any());
			verify(commandService).createCommand(requestCaptor.capture());
		});

		var request = requestCaptor.getValue();
		@SuppressWarnings("unchecked")
		Map<String, Object> item = (Map<String, Object>) request.getBodyObject().get("item");
		assertThat(item).containsEntry("text", "Es ist eine Fehler bei der Verarbeitung des Vorganges aufgetreten");
	}

	private Vorgang buildVorgang() {
		return Vorgang.builder()
				.id(VorgangId.from(VorgangTestFactory.ID))
				.formEngineName("testFormEngine")
				.formId("testForm")
				.build();
	}
}
