/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.HashMap;
import java.util.Map;

import jakarta.validation.ValidationException;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandRevokeFailedEvent;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.vorgang.command.CommandCreatedEventTestFactory;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangDeletedEventTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class VorgangAttachedItemEventListenerTest {

	@Spy
	@InjectMocks
	private VorgangAttachedItemEventListener listener;

	@Mock
	private VorgangAttachedItemService service;
	@Mock
	private ApplicationEventPublisher publisher;
	@Spy
	private VorgangAttachedItemMapper mapper = Mappers.getMapper(VorgangAttachedItemMapper.class);

	@Nested
	class TestCreateItem {

		private Map<String, Object> bodyMap = VorgangAttachedItemTestFactory.asMap();

		private Command command = CommandTestFactory.createBuilder().bodyObject(bodyMap).relationId(VorgangTestFactory.ID).build();

		@Captor
		private ArgumentCaptor<VorgangAttachedItem> itemCaptor;

		@Test
		void shouldThrowExceptionOnIdDismatching() {
			var command = CommandTestFactory.createBuilder().vorgangId("other").build();
			var event = CommandCreatedEventTestFactory.create(command);

			assertThatThrownBy(() -> listener.createItem(event)).isInstanceOf(ValidationException.class);

		}

		@DisplayName("should call service")
		@Nested
		class TestCallService {

			@Test
			void withItem() {
				listener.createItem(CommandCreatedEventTestFactory.create(command));

				verify(service).create(anyString(), itemCaptor.capture());
				assertThat(itemCaptor.getValue()).usingRecursiveComparison()
						.ignoringFields(VorgangAttachedItem.FIELDNAME_ID, VorgangAttachedItem.FIELDNAME_VERSION)
						.isEqualTo(VorgangAttachedItemTestFactory.create());
			}

			@Test
			void withIdAndVersionAsNull() {
				listener.createItem(CommandCreatedEventTestFactory.create(command));

				verify(service).create(anyString(), itemCaptor.capture());
				var item = itemCaptor.getValue();

				assertThat(item.getId()).isNull();
				assertThat(item.getVersion()).isZero();
			}

			@Test
			void withCommandId() {
				listener.createItem(CommandCreatedEventTestFactory.create(command));

				verify(service).create(eq(CommandTestFactory.ID), any());
			}

		}
	}

	@Nested
	class TestUpdateItem {
		private Map<String, Object> bodyMap = VorgangAttachedItemTestFactory.asMap();

		private Command command = buildCommand(bodyMap);

		@Captor
		private ArgumentCaptor<VorgangAttachedItem> itemCaptor;

		@DisplayName("service Call")
		@Nested
		class TestCallService {
			@Test
			void shouldHaveItem() {
				listener.updateItem(CommandCreatedEventTestFactory.create(command));

				verify(service).update(anyString(), itemCaptor.capture());
				assertThat(itemCaptor.getValue()).usingRecursiveComparison().isEqualTo(VorgangAttachedItemTestFactory.create());
			}

			@Test
			void shouldHaveCommand() {
				listener.updateItem(CommandCreatedEventTestFactory.create(command));

				verify(service).update(eq(CommandTestFactory.ID), any());
			}

			@Test
			void shouldHaveSetRelationIdAsId() {
				var bMap = new HashMap<>(bodyMap);
				bMap.put(VorgangAttachedItem.FIELDNAME_ID, "otherId");

				listener.updateItem(CommandCreatedEventTestFactory.create(buildCommand(bMap)));

				verify(service).update(anyString(), itemCaptor.capture());
				assertThat(itemCaptor.getValue().getId()).isEqualTo(VorgangAttachedItemTestFactory.ID);
			}
		}

		private Command buildCommand(Map<String, Object> bodyMap) {
			return CommandTestFactory.createBuilder().relationId(VorgangAttachedItemTestFactory.ID)
					.relationVersion(VorgangAttachedItemTestFactory.VERSION).bodyObject(bodyMap).build();
		}
	}

	@Nested
	class TestPatchItem {
		private Map<String, Object> bodyMap = Map.of("id", VorgangAttachedItemTestFactory.ID, "item", Map.of("number", "73"));

		private Command command = buildCommand(bodyMap);

		@DisplayName("service Call")
		@Nested
		class TestCallService {
			@SuppressWarnings("unchecked")
			@Test
			void shouldHaveItem() {
				listener.patchItem(CommandCreatedEventTestFactory.create(command));

				verify(service).patch(any(), (Map<String, Object>) eq(bodyMap.get("item")));
			}

			@Test
			void shouldHaveCommand() {
				listener.patchItem(CommandCreatedEventTestFactory.create(command));

				verify(service).patch(eq(command), any());
			}
		}

		private Command buildCommand(Map<String, Object> bodyMap) {
			return CommandTestFactory.createBuilder().relationId(VorgangAttachedItemTestFactory.ID)
					.relationVersion(VorgangAttachedItemTestFactory.VERSION).bodyObject(bodyMap).build();
		}
	}

	@Nested
	@DisplayName("Delete item")
	class TestDeleteItem {

		private Command command = CommandTestFactory.createBuilder().relationId(VorgangAttachedItemTestFactory.ID)
				.relationVersion(VorgangAttachedItemTestFactory.VERSION).build();

		@Nested
		@DisplayName("by itemId and version")
		class TestDeleteByIdAndVersion {

			@Test
			void shouldCallService() {
				listener.deleteItem(CommandCreatedEventTestFactory.create(command));

				verify(service).markAsDeleteByIdAndVersion(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);
			}

			@Test
			void shouldPublishCommandExecutedEvent() {
				listener.deleteItem(CommandCreatedEventTestFactory.create(command));

				verify(publisher).publishEvent(any(VorgangAttachedItemDeletedEvent.class));
			}

			@Test
			void shouldCallHandleException() {
				doThrow(new RuntimeException()).when(service).markAsDeleteByIdAndVersion(anyString(), anyLong());

				listener.deleteItem(CommandCreatedEventTestFactory.create(command));

				verify(listener).handleException(any(), eq(CommandTestFactory.ID));
			}
		}

		@Nested
		@DisplayName("by vorgangId")
		class TestDeleteByVorgangId {

			@Test
			void shouldCallService() {
				listener.onVorgangDeleted(VorgangDeletedEventTestFactory.create());

				verify(service).deleteByVorgangId(VorgangTestFactory.ID);
			}
		}
	}

	@Nested
	class TestRevokeCommand {

		@Nested
		class TestDeleteItem {

			@Test
			void shouldCallService() {
				listener.onRevokeDeleteItem(new RevokeCommandEvent(CommandTestFactory.createBuilder().build()));

				verify(service).unmarkAsDeleteByIdAndVersion(CommandTestFactory.RELATION_ID, CommandTestFactory.RELATION_VERSION);
			}

			@Test
			void shouldCallHandleException() {
				doThrow(new RuntimeException()).when(service).unmarkAsDeleteByIdAndVersion(anyString(), anyLong());

				listener.onRevokeDeleteItem(new RevokeCommandEvent(CommandTestFactory.createBuilder().build()));

				verify(publisher).publishEvent(any(CommandRevokeFailedEvent.class));
			}
		}
	}
}
