/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import static de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class VorgangAttachedItemServiceTest {

	@InjectMocks
	@Spy
	private VorgangAttachedItemService service;
	@Mock
	private VorgangAttachedItemRepository repository;
	@Mock
	private ApplicationEventPublisher publisher;

	@Nested
	class TestCreateItem {

		@Test
		void shouldCallRepository() {
			when(repository.save(any())).thenReturn(VorgangAttachedItemTestFactory.create());

			service.create(CommandTestFactory.ID, VorgangAttachedItemTestFactory.create());

			verify(repository).save(any());
		}

		@Test
		void shouldPublishEvent() {
			when(repository.save(any())).thenReturn(VorgangAttachedItemTestFactory.create());

			service.create(CommandTestFactory.ID, VorgangAttachedItemTestFactory.create());

			verify(publisher).publishEvent(any(VorgangAttachedItemCreatedEvent.class));
		}

		@Nested
		class TestPrepareForCreation {
			@Test
			void shouldRemoveId() {
				var prepared = service.prepareForCreation(VorgangAttachedItemTestFactory.create());

				assertThat(prepared.getId()).isNull();
			}

			@Test
			void shouldSetVersionToZero() {
				var prepared = service.prepareForCreation(VorgangAttachedItemTestFactory.create());

				assertThat(prepared.getVersion()).isZero();
			}

			@Test
			void shouldRemoveIdFromItem() {
				var prepared = service.prepareForCreation(VorgangAttachedItemTestFactory.create());

				assertThat(prepared.getItem()).doesNotContainKey(VorgangAttachedItem.FIELDNAME_ID);
			}

			@Test
			void shouldSetIsDeletedToFalse() {
				var prepared = service.prepareForCreation(VorgangAttachedItemTestFactory.create());

				assertThat(prepared.isDeleted()).isFalse();
			}
		}

	}

	@Nested
	class TestUpdateItem {

		@Test
		void shouldCallRepository() {
			service.update(CommandTestFactory.ID, VorgangAttachedItemTestFactory.create());

			verify(repository).patch(any(), anyLong(), any());
		}

		@Test
		void shouldPublishEvent() {
			service.update(CommandTestFactory.ID, VorgangAttachedItemTestFactory.create());

			verify(publisher).publishEvent(any(VorgangAttachedItemUpdatedEvent.class));
		}
	}

	@Nested
	class TestForceUpdateItem {

		@Test
		void shouldCallRepository() {
			service.forceUpdate(CommandTestFactory.ID, VorgangAttachedItemTestFactory.create());

			verify(repository).forcePatch(eq(VorgangAttachedItemTestFactory.ID), any());
		}

		@Test
		void shouldPublishEvent() {
			service.forceUpdate(CommandTestFactory.ID, VorgangAttachedItemTestFactory.create());

			verify(publisher).publishEvent(any(VorgangAttachedItemUpdatedEvent.class));
		}
	}

	@Nested
	class TestFindBy {

		@Test
		void shouldCallFindByIdAndDeleted() {
			service.findById(PostfachNachrichtTestFactory.ID);

			verify(repository).findByIdAndDeleted(PostfachNachrichtTestFactory.ID, VorgangAttachedItemService.NOT_DELETED);
		}

	}

	@Nested
	class TestFind {

		@Test
		void shouldCallFindByClientAndItemNameAndVorgangIdAndDeleted() {
			service.find(VorgangTestFactory.ID, Optional.of(CLIENT), Optional.of(ITEM_NAME));

			verify(repository).findByVorgangIdAndClientAndItemNameAndDeleted(VorgangTestFactory.ID, CLIENT, ITEM_NAME,
					VorgangAttachedItemService.NOT_DELETED);
		}

		@Test
		void shouldCallFindByClientAndVorgangIdAndDeleted() {
			service.find(VorgangTestFactory.ID, Optional.of(CLIENT), Optional.empty());

			verify(repository).findByVorgangIdAndClientAndDeleted(VorgangTestFactory.ID, CLIENT, VorgangAttachedItemService.NOT_DELETED);
		}

		@Test
		void shouldCallFindByItemNameAndVorgangIdAndDeleted() {
			service.find(VorgangTestFactory.ID, Optional.empty(), Optional.of(ITEM_NAME));

			verify(repository).findByVorgangIdAndItemNameAndDeleted(VorgangTestFactory.ID, ITEM_NAME, VorgangAttachedItemService.NOT_DELETED);
		}

		@Test
		void shouldCallFindByVorgangIdAndDeleted() {
			service.find(VorgangTestFactory.ID, Optional.empty(), Optional.empty());

			verify(repository).findByVorgangIdAndDeleted(VorgangTestFactory.ID, VorgangAttachedItemService.NOT_DELETED);
		}
	}

	@Nested
	class TestPatch {

		@Test
		void shouldCallRepository() {
			Map<String, Object> propertyMap = Map.of(VorgangAttachedItemTestFactory.ITEM_FIELD_NAME,
					VorgangAttachedItemTestFactory.ITEM_FIELD_INT_VALUE);
			var item = VorgangAttachedItemTestFactory.createBuilder().item(propertyMap).build();
			var command = buildCommand(propertyMap);

			service.patch(command, item.getItem());

			verify(repository).patch(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION, propertyMap);
		}

		private Command buildCommand(Map<String, Object> bodyMap) {
			return CommandTestFactory.createBuilder()
					.relationId(VorgangAttachedItemTestFactory.ID)
					.relationVersion(VorgangAttachedItemTestFactory.VERSION)
					.bodyObject(bodyMap).build();
		}
	}

	@Nested
	class TestForcePatchItem {

		@Test
		void shouldCallRepository() {
			Map<String, Object> fieldValueMap = Map.of(VorgangAttachedItemTestFactory.ITEM_FIELD_NAME,
					VorgangAttachedItemTestFactory.ITEM_FIELD_INT_VALUE);

			service.forcePatch(VorgangAttachedItemTestFactory.ID, fieldValueMap);

			verify(repository).forcePatch(VorgangAttachedItemTestFactory.ID, fieldValueMap);
		}
	}

	@Nested

	class TestGetById {
		@Test
		void shouldReturnAttachedItem() {
			var item = VorgangAttachedItemTestFactory.create();
			when(repository.findByIdAndDeleted(any(), anyBoolean())).thenReturn(Optional.of(item));

			var attachedItem = service.getById(VorgangAttachedItemTestFactory.ID);

			assertThat(attachedItem).usingRecursiveComparison().ignoringFields(VorgangAttachedItem.FIELDNAME_ITEM).isEqualTo(item);
		}

		@DisplayName("should throw an exception if no attachedItem was found")
		@Test
		void shouldThrowException() {
			when(repository.findByIdAndDeleted(any(), anyBoolean())).thenReturn(Optional.empty());

			assertThrows(NotFoundException.class, () -> service.getById(VorgangAttachedItemTestFactory.ID));
		}

		@Test
		void shouldCallFindByIdAndDeleted() {
			var item = VorgangAttachedItemTestFactory.create();
			when(repository.findByIdAndDeleted(any(), anyBoolean())).thenReturn(Optional.of(item));

			service.getById(VorgangAttachedItemTestFactory.ID);

			verify(repository).findByIdAndDeleted(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemService.NOT_DELETED);
		}
	}

	@Test
	void shouldAddIdToItem() {
		String id = UUID.randomUUID().toString();

		var enrichedItem = service.addIdToItem(VorgangAttachedItemTestFactory.createBuilder().id(id).build());

		assertThat(enrichedItem.getItem()).containsEntry(VorgangAttachedItem.FIELDNAME_ID, id);
	}

	@Nested
	@DisplayName("Set deleted by ID and version")
	class TestMarkDeletedByIdAndVersion {

		@Test
		@DisplayName("should call updateDeletedByIdAndVersion to mark item as deleted")
		void shouldCallUpdateIsDeletedMethod() {
			var deletedFlag = true;

			service.markAsDeleteByIdAndVersion(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);

			verify(repository).updateDeleted(deletedFlag, VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);
		}

		@Test
		@DisplayName("should call updateDeletedByIdAndVersion to mark item as not deleted")
		void shouldCallUpdateIsDeletedMethod1() {
			var deletedFlag = false;

			service.unmarkAsDeleteByIdAndVersion(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);

			verify(repository).updateDeleted(deletedFlag, VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);
		}
	}

	@Nested
	@DisplayName("Delete item by vorgangId")
	class TestDelete {

		@Test
		void shouldCallDeleteMethod() {
			service.deleteByVorgangId(VorgangAttachedItemTestFactory.ID);

			verify(repository).deleteByVorgangId(VorgangAttachedItemTestFactory.ID);
		}

	}
}
