/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Stream;

import org.apache.http.entity.ContentType;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.MessageAttachmentTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemMapper;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemService;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.callcontext.UserTestFactory;
import de.ozgcloud.vorgang.files.FileId;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.files.OzgFile;
import de.ozgcloud.vorgang.files.UploadedFilesReference;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class PersistPostfachMailByCommandServiceTest {

	@InjectMocks
	private PersistPostfachNachrichtByCommandService service;
	@Mock
	private CommandService commandService;
	@Mock
	private VorgangAttachedItemService attachedItemService;
	@Mock
	private FileService fileService;

	@DisplayName("Persist mail")
	@Nested
	class TestPersistMail {

		@Captor
		private ArgumentCaptor<CreateCommandRequest> requestCaptor;

		@Test
		void shouldCallCommandService() {
			callService();

			verify(commandService).createCommand(notNull());
		}

		@Test
		void shouldContainVorgangId() {
			var request = callService();

			assertThat(request.getVorgangId()).isEqualTo(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldContainRelationInformations() {
			var request = callService();

			assertThat(request.getRelationId()).isEqualTo(MessageTestFactory.VORGANG_ID);
			assertThat(request.getRelationVersion()).isNull();
		}

		@Test
		void shouldHaveOrder() {
			var request = callService();

			assertThat(request.getOrder()).isEqualTo(Order.CREATE_ATTACHED_ITEM.name());
		}

		@Test
		void shouldHaveCallContext() {
			var request = callService();

			assertThat(request.getCallContext()).isNotNull();
			assertThat(request.getCallContext().getUser().getId()).isEqualTo(UserTestFactory.ID);
		}

		@Test
		void shouldHaveVorgangAttachedItem() {
			var request = callService();

			assertThat(request.getBodyObject())
					.containsEntry(VorgangAttachedItemMapper.PROPERTY_VORGANG_ID, MessageTestFactory.VORGANG_ID)
					.containsEntry(VorgangAttachedItemMapper.PROPERTY_CLIENT, "OzgCloud_NachrichtenManager")
					.containsEntry(VorgangAttachedItemMapper.PROPERTY_ITEM_NAME, "PostfachMail")
					.containsKey(VorgangAttachedItemMapper.PROPERTY_ITEM);
		}

		@Test
		void shouldNotAddUser() {
			service.persistNachricht(Optional.empty(), PostfachNachrichtTestFactory.create());
			verify(commandService).createCommand(requestCaptor.capture());

			assertThat(requestCaptor.getValue().getCallContext().getUser()).isNull();
		}

		private CreateCommandRequest callService() {
			service.persistNachricht(Optional.of(UserTestFactory.ID), PostfachNachrichtTestFactory.createBuilder().id(null).build());
			verify(commandService).createCommand(requestCaptor.capture());

			return requestCaptor.getValue();
		}

		@DisplayName("build mail map")
		@Nested
		class TestBuildMailMap {

			@Test
			void shouldContainAllFieldsIgnoringId() {
				var map = service.buildNachrichtMap(PostfachNachrichtTestFactory.createBuilder().id(null).build());

				var mapWithoutId = PostfachNachrichtTestFactory.asMap();
				mapWithoutId.remove("id");

				assertThat(map).containsAllEntriesOf(mapWithoutId);
			}

			@Test
			void shouldIgnoreEmptyMessageId() {
				assertDoesNotThrow(() -> service.buildNachrichtMap(PostfachNachrichtTestFactory.createBuilder().messageId(null).build()));
			}

			@Test
			void shouldIgnoreEmptyCreatedBy() {
				assertDoesNotThrow(() -> service.buildNachrichtMap(PostfachNachrichtTestFactory.createBuilder().createdBy(null).build()));
			}

			@Test
			void shouldHandleNullAsPostfachId() {
				var postfachMail = PostfachNachrichtTestFactory.createBuilder().postfachId(null).build();

				assertDoesNotThrow(() -> service.buildNachrichtMap(postfachMail));
			}

			@Test
			void shouldProceedWithEmptyPostfachAddress() {
				var postfachMail = PostfachNachrichtTestFactory.createBuilder().postfachAddress(null).build();

				assertDoesNotThrow(() -> service.buildNachrichtMap(postfachMail));
			}
		}
	}

	@Nested
	class TestFindById {

		@BeforeEach
		void init() {
			when(attachedItemService.findById(anyString()))
					.thenReturn(Optional.of(VorgangAttachedItemTestFactory.createBuilder().item(PostfachNachrichtTestFactory.asMap()).build()));
		}

		@Test
		void shouldCallItemService() {
			service.findById(PostfachNachrichtTestFactory.ID);

			verify(attachedItemService).findById(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldReturnNachrichtAsMap() {
			var found = service.findById(PostfachNachrichtTestFactory.ID);

			assertThat(found).isPresent().get().usingRecursiveComparison().isEqualTo(PostfachNachrichtTestFactory.asMap());
		}
	}

	@Nested
	class TestFindByVorgang {

		@BeforeEach
		void init() {
			when(attachedItemService.find(anyString(), any(), any()))
					.thenReturn(Stream.of(VorgangAttachedItemTestFactory.createBuilder().item(PostfachNachrichtTestFactory.asMap()).build()));
		}

		@Test
		void shouldCallItemService() {
			service.findByVorgangAsMap(VorgangTestFactory.ID);

			verify(attachedItemService).find(VorgangTestFactory.ID, Optional.of(PersistPostfachNachrichtByCommandService.CLIENT),
					Optional.of(PersistPostfachNachrichtByCommandService.ITEM_NAME));
		}

		@Test
		void shouldReturnNachrichtenAsMap() {
			var found = service.findByVorgangAsMap(VorgangTestFactory.ID);

			assertThat(found).first().usingRecursiveComparison().isEqualTo(PostfachNachrichtTestFactory.asMap());
		}
	}

	@Nested
	class TestPatch {

		@Test
		void shoulCallRepository() {
			Map<String, Object> propertyMap = Map.of(VorgangAttachedItemTestFactory.ITEM_FIELD_NAME,
					VorgangAttachedItemTestFactory.ITEM_FIELD_INT_VALUE);

			service.patch(VorgangAttachedItemTestFactory.ID, propertyMap);

			verify(attachedItemService).forcePatch(VorgangAttachedItemTestFactory.ID, propertyMap);
		}
	}

	@Nested
	class TestGetById {

		@BeforeEach
		void mockService() {
			when(attachedItemService.getById(anyString()))
					.thenReturn(VorgangAttachedItemTestFactory.createBuilder().item(PostfachNachrichtTestFactory.asMap()).build());
		}

		@Test
		void shouldCallService() {
			service.getById(PostfachNachrichtTestFactory.ID);

			verify(attachedItemService).getById(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldReturnItem() {
			Map<String, Object> item = service.getById(PostfachNachrichtTestFactory.ID);

			assertThat(item).isEqualTo(PostfachNachrichtTestFactory.asMap());
		}
	}

	@Nested
	class TestUploadedFilesReferenceCreation {
		@Test
		void shouldCreateUploadedFilesReference() {
			UploadedFilesReference ref = service.createUploadedFilesReference(MessageTestFactory.VORGANG_ID);

			assertThat(ref.getVorgangId()).isEqualTo(MessageTestFactory.VORGANG_ID);
			assertThat(ref.getClient()).isEqualTo(PersistPostfachNachrichtByCommandService.CLIENT);
			assertThat(ref.getName()).isEqualTo(PersistPostfachNachrichtByCommandService.ATTACHMENT_NAME);
		}
	}

	@Nested
	class TestOzgFileCreation {
		@Test
		void shouldCreateOzgFile() {
			OzgFile file = service.createOzgFile(MessageAttachmentTestFactory.FILENAME, ContentType.APPLICATION_OCTET_STREAM.toString(),
					MessageAttachmentTestFactory.SIZE);

			assertThat(file.getContentType()).isEqualTo(ContentType.APPLICATION_OCTET_STREAM.toString());
			assertThat(file.getSize()).isEqualTo(MessageAttachmentTestFactory.SIZE);
			assertThat(file.getName()).isEqualTo(MessageAttachmentTestFactory.FILENAME);
		}
	}

	@DisplayName("On Persisting Attachement")
	@Nested
	class TestPersistAttachment {
		@BeforeEach
		void init() {
			when(fileService.uploadFileStream(any(), any(), any(), any())).thenReturn(CompletableFuture.completedFuture(FileId.from("42")));
		}

		@DisplayName("Upload files Reference")
		@Nested
		class TestUploadFilesReference {

			@Captor
			private ArgumentCaptor<UploadedFilesReference> refCaptor;

			@Test
			void shouldHaveClientName() {
				var ref = callService();

				assertThat(ref.getClient()).isEqualTo(PersistPostfachNachrichtByCommandService.CLIENT);
			}

			@Test
			void shouldHaveName() {
				var ref = callService();

				assertThat(ref.getName()).isEqualTo(PersistPostfachNachrichtByCommandService.ATTACHMENT_NAME);
			}

			@Test
			void shouldHaveVorgangId() {
				var ref = callService();

				assertThat(ref.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
			}

			private UploadedFilesReference callService() {
				service.persistAttachment(VorgangTestFactory.ID, AttachmentFileTestFactory.create());

				verify(fileService).uploadFileStream(refCaptor.capture(), any(), any(), any());

				return refCaptor.getValue();
			}
		}

		@DisplayName("OZG-File")
		@Nested
		class TestOzgFile {
			@Captor
			private ArgumentCaptor<OzgFile> ozgFileCaptor;

			@Test
			void shouldHaveName() {
				var file = callService();

				assertThat(file.getName()).isEqualTo(MessageAttachmentTestFactory.FILENAME);
			}

			@Test
			void shouldHaveSize() {
				var file = callService();

				assertThat(file.getSize()).isEqualTo(MessageAttachmentTestFactory.SIZE);
			}

			@Test
			void shouldHaveContentType() {
				var file = callService();

				assertThat(file.getContentType()).isEqualTo("text/plain");
			}

			private OzgFile callService() {
				var attachmentFile = AttachmentFileTestFactory.createBuilder().content(AttachmentFileTestFactory.BASE64_CONTENT_SUPPLIER).build();

				service.persistAttachment(VorgangTestFactory.ID, attachmentFile);

				verify(fileService).uploadFileStream(any(), ozgFileCaptor.capture(), any(), any());

				return ozgFileCaptor.getValue();
			}
		}

		@Test
		void shouldHaveContent() {
			ArgumentCaptor<InputStream> contentCaptor = ArgumentCaptor.forClass(InputStream.class);
			var attachment = AttachmentFileTestFactory.createBuilder().content(AttachmentFileTestFactory.BASE64_CONTENT_SUPPLIER).build();

			service.persistAttachment(VorgangTestFactory.ID, attachment);

			verify(fileService).uploadFileStream(any(), any(), any(), contentCaptor.capture());
			assertThat(contentCaptor.getValue()).hasContent(AttachmentFileTestFactory.CONTENT);
		}

		@Test
		void shouldReturnFileId() {
			var id = service.persistAttachment(VorgangTestFactory.ID, AttachmentFileTestFactory.create());

			assertThat(id).isEqualTo("42");
		}
	}

	@Nested
	class TestGetTypeByFileName {
		@Test
		void shouldReturnDocxType() {
			var attachmentFile = AttachmentFileTestFactory.createBuilder().name("file.docx").contentType("word-file").build();

			var type = service.getTypeByFile(attachmentFile);

			assertThat(type).isEqualTo("application/vnd.openxmlformats-officedocument.wordprocessingml.document");
		}

		@Test
		void shouldReturnOctedStreamAsUnkown() {
			var attachmentFile = AttachmentFileTestFactory.createBuilder().name("file.qtsch").contentType("lalal").build();

			var type = service.getTypeByFile(attachmentFile);

			assertThat(type).isEqualTo(ContentType.APPLICATION_OCTET_STREAM.toString());
		}

		@Test
		void shouldReturnOdtType() {
			var attachmentFile = AttachmentFileTestFactory.createBuilder().name("file.odt").contentType("opendoc file").build();

			var type = service.getTypeByFile(attachmentFile);

			assertThat(type).isEqualTo("application/vnd.oasis.opendocument.text");
		}
	}
}
